import sys
import odak
from odak import np

def test():
    px_size           = 30./151.
    dist              = 580. # 58 cm
    angsize           = np.degrees(np.arctan(px_size/dist))    
    image             = np.random.rand(2000,2000)
    location          = [340,360,1010,1050]
    line_x,line_y,roi = odak.measurement.roi(
                                         image,
                                         location=location,
                                        )
    mtf,freq          = odak.measurement.modulation_transfer_function(
                                                                      line_x,
                                                                      line_y,
                                                                      px_size=[angsize,angsize],
                                                                     )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



import sys

def test():
    from odak import np
    import torch
    from odak.learn.wave import gerchberg_saxton,produce_phase_only_slm_pattern,calculate_amplitude
    from odak.tools import save_image
    wavelength             = 0.000000532
    dx                     = 0.0000064
    distance               = 0.2
    input_field            = np.zeros((500,500),dtype=np.complex64)
    input_field[0::50,:]  += 1
    iteration_number       = 3
    if np.__name__ == 'cupy':
        input_field = np.asnumpy(input_field)
    input_field            = torch.from_numpy(input_field)
    hologram,reconstructed = gerchberg_saxton(
                                              input_field,
                                              iteration_number,
                                              distance,
                                              dx,
                                              wavelength,
                                              np.pi*2,
                                              'TR Fresnel'
                                             )
    # hologram,_             = produce_phase_only_slm_pattern(
    #                                                         hologram,
    #                                                         2*np.pi
    #                                                         )
    # amplitude              = calculate_amplitude(reconstructed)
    # amplitude              = amplitude.numpy()
    # save_image(
    #             'output_amplitude_torch.png',
    #             amplitude,
    #             cmin=0,
    #             cmax=np.amax(amplitude)
    #             )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



import sys
import os
import odak
from odak import np
from odak.wave import wavenumber,propagate_beam,add_random_phase

def main():
    # Variables to be set.
    wavelength                 = 0.5*pow(10,-6)
    pixeltom                   = 6*pow(10,-6)
    distance                   = 10.0
    propagation_type           = 'Fraunhofer'
    k                          = wavenumber(wavelength)
    sample_field               = np.zeros((150,150),dtype=np.complex64)
    sample_field               = np.zeros((150,150),dtype=np.complex64)
    sample_field[
                 65:85,
                 65:85
                ]              = 1
    sample_field               = add_random_phase(sample_field)
    hologram                   = propagate_beam(
                                                sample_field,
                                                k,
                                                distance,
                                                pixeltom,
                                                wavelength,
                                                propagation_type
                                               )
    if propagation_type == 'Fraunhofer':
        # Uncomment if you want to match the physical size of hologram and input field.
        #from odak.wave import fraunhofer_equal_size_adjust
        #hologram         = fraunhofer_equal_size_adjust(hologram,distance,pixeltom,wavelength)
        propagation_type = 'Fraunhofer Inverse'
        distance = np.abs(distance)
    reconstruction             = propagate_beam(
                                                hologram,
                                                k,
                                                -distance,
                                                pixeltom,
                                                wavelength,
                                                propagation_type
                                               )

    #from odak.visualize.plotly import detectorshow
    #detector       = detectorshow()
    #detector.add_field(sample_field)
    #detector.show()
    #detector       = detectorshow()
    #detector.add_field(hologram)
    #detector.show()
    #detector       = detectorshow()
    #detector.add_field(reconstruction)
    #detector.show()
    assert True==True

if __name__ == '__main__':
    sys.exit(main())



from odak import np
import pkg_resources
import finufft

def nufft2(field,fx,fy,size=None,sign=1,eps=10**(-12)):
    """
    A definition to take 2D Non-Uniform Fast Fourier Transform (NUFFT).

    Parameters
    ----------
    field       : ndarray
                  Input field.
    fx          : ndarray
                  Frequencies along x axis.
    fy          : ndarray
                  Frequencies along y axis.
    size        : list
                  Size.
    sign        : float
                  Sign of the exponential used in NUFFT kernel.
    eps         : float
                  Accuracy of NUFFT.

    Returns
    ----------
    result      : ndarray
                  Inverse NUFFT of the input field.
    """
    if np.__name__ == 'cupy':
        fx    = np.asnumpy(fx).astype(np.float64)
        fy    = np.asnumpy(fy).astype(np.float64)
        image = np.asnumpy(np.copy(field)).astype(np.complex128)
    else:
        image = np.copy(field).astype(np.complex128)
    result = finufft.nufft2d2(fx.flatten(),fy.flatten(),image,eps=eps,isign=sign)
    if type(size) == type(None):
        result = result.reshape(field.shape)
    else:
        result = result.reshape(size)
    if np.__name__ == 'cupy':
        result = np.asarray(result)
    return result

def nuifft2(field,fx,fy,size=None,sign=1,eps=10**(-12)):
    """
    A definition to take 2D Adjoint Non-Uniform Fast Fourier Transform (NUFFT).

    Parameters
    ----------
    field       : ndarray
                  Input field.
    fx          : ndarray
                  Frequencies along x axis.
    fy          : ndarray
                  Frequencies along y axis.
    size        : list or ndarray
                  Shape of the NUFFT calculated for an input field.
    sign        : float
                  Sign of the exponential used in NUFFT kernel.
    eps         : float
                  Accuracy of NUFFT.

    Returns
    ----------
    result      : ndarray
                  NUFFT of the input field.
    """
    if np.__name__ == 'cupy':
        fx    = np.asnumpy(fx).astype(np.float64)
        fy    = np.asnumpy(fy).astype(np.float64)
        image = np.asnumpy(np.copy(field)).astype(np.complex128)
    else:
        image = np.copy(field).astype(np.complex128)
    if type(size) == type(None):
        result = finufft.nufft2d1(
                                  fx.flatten(),
                                  fy.flatten(),
                                  image.flatten(),
                                  image.shape,
                                  eps=eps,
                                  isign=sign
                                 )
    else:
        result = finufft.nufft2d1(
                                  fx.flatten(),
                                  fy.flatten(),
                                  image.flatten(),
                                  (size[0],size[1]),
                                  eps=eps,
                                  isign=sign
                                 )
    if np.__name__ == 'cupy':
        result = np.asarray(result)
    return result

def generate_bandlimits(size=[512,512],levels=9):
    """
    A definition to calculate octaves used in bandlimiting frequencies in the frequency domain.

    Parameters
    ----------
    size       : list
                 Size of each mask in octaves.

    Returns
    ----------
    masks      : ndarray
                 Masks (Octaves).
    """
    masks = np.zeros((levels,size[0],size[1]))
    cx     = int(size[0]/2)
    cy     = int(size[1]/2)
    for i in range(0,masks.shape[0]):
        deltax = int((size[0])/(2**(i+1)))
        deltay = int((size[1])/(2**(i+1)))
        masks[
              i,
              cx-deltax:cx+deltax,
              cy-deltay:cy+deltay
             ] = 1.
        masks[
              i,
              int(cx-deltax/2.):int(cx+deltax/2.),
              int(cy-deltay/2.):int(cy+deltay/2.)
             ] = 0.
    masks = np.asarray(masks)
    return masks

def zero_pad(field,size=None,method='center'):
    """
    Definition to zero pad a MxN array to 2Mx2N array.

    Parameters
    ----------
    field             : ndarray
                        Input field MxN array.
    size              : list
                        Size to be zeropadded.
    method            : str
                        Zeropad either by placing the content to center or to the left.

    Returns
    ----------
    field_zero_padded : ndarray
                        Zeropadded version of the input field.
    """
    if type(size) == type(None):
        hx = int(np.ceil(field.shape[0])/2)
        hy = int(np.ceil(field.shape[1])/2)
    else:
        hx = int(np.ceil((size[0]-field.shape[0])/2))
        hy = int(np.ceil((size[1]-field.shape[1])/2))
    if method == 'center':
        field_zero_padded = np.pad(field,([hx,hx],[hy,hy]), constant_values=(0,0))
    elif method == 'left aligned':
        field_zero_padded = np.pad(field,([0,2*hx],[0,2*hy]), constant_values=(0,0))
    if type(size) != type(None):
        field_zero_padded = field_zero_padded[0:size[0],0:size[1]]
    return field_zero_padded

def crop_center(field,size=None):
    """
    Definition to crop the center of a field with 2Mx2N size. The outcome is a MxN array.

    Parameters
    ----------
    field       : ndarray
                  Input field 2Mx2N array.

    Returns
    ----------
    cropped     : ndarray
                  Cropped version of the input field.
    """
    if type(size) == type(None):
        qx      = int(np.ceil(field.shape[0])/4)
        qy      = int(np.ceil(field.shape[1])/4)
        cropped = np.copy(field[qx:3*qx,qy:3*qy])
    else:
        cx      = int(np.ceil(field.shape[0]/2))
        cy      = int(np.ceil(field.shape[1]/2))
        hx      = int(np.ceil(size[0]/2))
        hy      = int(np.ceil(size[1]/2))
        cropped = np.copy(field[cx-hx:cx+hx,cy-hy:cy+hy]) 
    return cropped

def quantize(image_field,bits=4):
    """
    Definitio to quantize a image field (0-255, 8 bit) to a certain bits level.

    Parameters
    ----------
    image_field : ndarray
                  Input image field.
    bits        : int
                  A value in between 0 to 8. Can not be zero.

    Returns
    ----------
    new_field   : ndarray
                  Quantized image field.
    """
    divider   = 2**(8-bits)
    new_field = image_field/divider
    new_field = new_field.astype(np.int)
    return new_field

def convolve2d(field,kernel):
    """
    Definition to convolve a field with a kernel by multiplying in frequency space.

    Parameters
    ----------
    field       : ndarray
                  Input field with MxN shape.
    kernel      : ndarray
                  Input kernel with MxN shape.

    Returns
    ----------
    new_field   : ndarray
                  Convolved field.
    """
    fr        = np.fft.fft2(field)
    fr2       = np.fft.fft2(np.flipud(np.fliplr(kernel)))
    m,n       = fr.shape
    new_field = np.real(np.fft.ifft2(fr*fr2))
    new_field = np.roll(new_field,int(-m/2+1),axis=0)
    new_field = np.roll(new_field,int(-n/2+1),axis=1)
    return new_field

def generate_2d_gaussian(kernel_length=[21,21], nsigma=[3,3]):
    """
    Generate 2D Gaussian kernel. Inspired from https://stackoverflow.com/questions/29731726/how-to-calculate-a-gaussian-kernel-matrix-efficiently-in-numpy

    Parameters
    ----------
    kernel_length : list
                    Length of the Gaussian kernel along X and Y axes.
    nsigma        : list
                    Sigma of the Gaussian kernel along X and Y axes.
    
    Returns
    ----------
    kernel_2d     : ndarray
                    Generated Gaussian kernel.
    """
    x           = np.linspace(-nsigma[0], nsigma[0], kernel_length[0]+1)
    y           = np.linspace(-nsigma[1], nsigma[1], kernel_length[1]+1)
    xx, yy      = np.meshgrid(x, y)
    kernel_2d   = np.exp(-0.5*(np.square(xx)/np.square(nsigma[0]) + np.square(yy)/np.square(nsigma[1])))
    kernel_2d   = kernel_2d/kernel_2d.sum()
    return kernel_2d

def blur_gaussian(field,kernel_length=[21,21],nsigma=[3,3]):
    """
    A definition to blur a field using a Gaussian kernel.

    Parameters
    ----------
    field         : ndarray
                    MxN field.
    kernel_length : list
                    Length of the Gaussian kernel along X and Y axes.
    nsigma        : list
                    Sigma of the Gaussian kernel along X and Y axes.

    Returns
    ----------
    blurred_field : ndarray
                    Blurred field.
    """
    kernel        = generate_2d_gaussian(kernel_length,nsigma)
    kernel        = zero_pad(kernel,field.shape)
    blurred_field = convolve2d(field,kernel)
    blurred_field = blurred_field/np.amax(blurred_field)
    return blurred_field



from odak import np
from odak.tools import nufft2,nuifft2,zero_pad
from .lens import quadratic_phase_function
from .__init__ import wavenumber,calculate_amplitude,calculate_phase,set_amplitude,generate_complex_field,add_random_phase,add_phase
from tqdm import tqdm

def propagate_beam(field,k,distance,dx,wavelength,propagation_type='IR Fresnel'):
    """
    Definitions for Fresnel Impulse Respone (IR), Angular Spectrum (AS), Bandlimited Angular Spectrum (BAS), Fresnel Transfer Function (TF), Fraunhofer diffraction in accordence with "Computational Fourier Optics" by David Vuelz. For more on Bandlimited Fresnel impulse response also known as Bandlimited Angular Spectrum method see "Band-limited Angular Spectrum Method for Numerical Simulation of Free-Space Propagation in Far and Near Fields".

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.
    propagation_type : str
                       Type of the propagation (IR Fresnel, Angular Spectrum, Bandlimited Angular Spectrum, TR Fresnel, Fraunhofer).

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    if propagation_type == 'Rayleigh-Sommerfeld':
        result = rayleigh_sommerfeld(field,k,distance,dx,wavelength)
    elif propagation_type == 'Angular Spectrum':
        result = angular_spectrum(field,k,distance,dx,wavelength)
    elif propagation_type == 'IR Fresnel':
        result = impulse_response_fresnel(field,k,distance,dx,wavelength)
    elif propagation_type == 'Bandlimited Angular Spectrum':
        result = band_limited_angular_spectrum(field,k,distance,dx,wavelength)
    elif propagation_type == 'Bandextended Angular Spectrum':
        result = band_extended_angular_spectrum(field,k,distance,dx,wavelength)
    elif propagation_type == 'Adaptive Sampling Angular Spectrum':
        result = adaptive_sampling_angular_spectrum(field,k,distance,dx,wavelength)
    elif propagation_type == 'TR Fresnel':
        result = transfer_function_fresnel(field,k,distance,dx,wavelength)
    elif propagation_type == 'Fraunhofer':
        result = fraunhofer(field,k,distance,dx,wavelength)
    elif propagation_type == 'Fraunhofer Inverse':
        result = fraunhofer_inverse(field,k,distance,dx,wavelength)
    else:
        raise Exception("Unknown propagation type selected.")
    return result

def adaptive_sampling_angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate adaptive sampling angular spectrum based beam propagation. For more Zhang, Wenhui, Hao Zhang, and Guofan Jin. "Adaptive-sampling angular spectrum method with full utilization of space-bandwidth product." Optics Letters 45.16 (2020): 4416-4419.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    iflag = -1
    eps   = 10**(-12)
    nv,nu = field.shape
    l     = nu*dx
    x     = np.linspace(-l/2,l/2,nu)
    y     = np.linspace(-l/2,l/2,nv)
    X,Y   = np.meshgrid(x,y)
    fx    = np.linspace(-1./2./dx,1./2./dx,nu)
    fy    = np.linspace(-1./2./dx,1./2./dx,nv)
    FX,FY = np.meshgrid(fx,fy)
    forig = 1./2./dx
    fc2   = 1./2*(nu/wavelength/np.abs(distance))**0.5
    ss    = np.abs(fc2)/forig
    zc    = nu*dx**2/wavelength
    K     = nu/2/np.amax(np.abs(fx))
    m     = 2
    nnu2  = m*nu
    nnv2  = m*nv
    fxn   = np.linspace(-1./2./dx,1./2./dx,nnu2)
    fyn   = np.linspace(-1./2./dx,1./2./dx,nnv2)
    if np.abs(distance) > zc*2:
        fxn   = fxn*ss
        fyn   = fyn*ss
    FXN,FYN   = np.meshgrid(fxn,fyn)
    Hn        = np.exp(1j*k*distance*(1-(FXN*wavelength)**2-(FYN*wavelength)**2)**0.5)
    FX        = FXN/np.amax(FXN)*np.pi
    FY        = FYN/np.amax(FYN)*np.pi
    t_2       = nufft2(field,FX*ss,FY*ss,size=[nnv2,nnu2],sign=iflag,eps=eps)
    FX        = FX/np.amax(FX)*np.pi
    FY        = FY/np.amax(FY)*np.pi
    result    = nuifft2(Hn*t_2,FX*ss,FY*ss,size=[nv,nu],sign=-iflag,eps=eps)
    return result

def fraunhofer_equal_size_adjust(field,distance,dx,wavelength):
    """
    A definition to match the physical size of the original field with the propagated field.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    new_field        : np.complex
                       Final complex field (MxN).
    """
    nv,nu     = field.shape
    l1        = nu*dx
    l2        = wavelength*distance/dx
    m         = l1/l2
    px        = int(m*nu)
    py        = int(m*nv)
    nx        = int(field.shape[0]/2-px/2)
    ny        = int(field.shape[1]/2-py/2)
    new_field = np.copy(field[nx:nx+px,ny:ny+py])
    return new_field

def fraunhofer(field,k,distance,dx,wavelength):
    """
    A definition to calculate Fraunhofer based beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    nv,nu  = field.shape
    l      = nu*dx
    l2     = wavelength*distance/dx
    dx2    = wavelength*distance/l
    fx     = np.linspace(-l2/2.,l2/2.,nu)
    fy     = np.linspace(-l2/2.,l2/2.,nv)
    FX,FY  = np.meshgrid(fx,fy)
    FZ     = FX**2+FY**2
    c      = np.exp(1j*k*distance)/(1j*wavelength*distance)*np.exp(1j*k/(2*distance)*FZ)
    result = c*np.fft.ifftshift(np.fft.fft2(np.fft.fftshift(field)))*dx**2
    return result

def fraunhofer_inverse(field,k,distance,dx,wavelength):
    """
    A definition to calculate Inverse Fraunhofer based beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    distance = np.abs(distance)
    nv,nu    = field.shape
    l        = nu*dx
    l2       = wavelength*distance/dx
    dx2      = wavelength*distance/l
    fx       = np.linspace(-l2/2.,l2/2.,nu)
    fy       = np.linspace(-l2/2.,l2/2.,nv)
    FX,FY    = np.meshgrid(fx,fy)
    FZ       = FX**2+FY**2
    c        = np.exp(1j*k*distance)/(1j*wavelength*distance)*np.exp(1j*k/(2*distance)*FZ)
    result   = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(field/dx**2/c )))
    return result

def band_limited_angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate bandlimited angular spectrum based beam propagation. For more Matsushima, Kyoji, and Tomoyoshi Shimobaba. "Band-limited angular spectrum method for numerical simulation of free-space propagation in far and near fields." Optics express 17.22 (2009): 19662-19673.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    nv,nu  = field.shape
    x      = np.linspace(-nu/2*dx,nu/2*dx,nu)
    y      = np.linspace(-nv/2*dx,nv/2*dx,nv)
    X,Y    = np.meshgrid(x,y)
    Z      = X**2+Y**2
    h      = 1./(1j*wavelength*distance)*np.exp(1j*k*(distance+Z/2/distance))
    h      = np.fft.fft2(np.fft.fftshift(h))*dx**2
    flimx  = np.ceil(1/(((2*distance*(1./(nu)))**2+1)**0.5*wavelength))
    flimy  = np.ceil(1/(((2*distance*(1./(nv)))**2+1)**0.5*wavelength))
    mask   = np.zeros((nu,nv),dtype=np.complex64)
    mask   = (np.abs(X)<flimx) & (np.abs(Y)<flimy)
    mask   = set_amplitude(h,mask)
    U1     = np.fft.fft2(np.fft.fftshift(field))
    U2     = mask*U1
    result = np.fft.ifftshift(np.fft.ifft2(U2))
    return result

def angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate angular spectrum based beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    nv,nu  = field.shape
    x      = np.linspace(-nu/2*dx,nu/2*dx,nu)
    y      = np.linspace(-nv/2*dx,nv/2*dx,nv)
    X,Y    = np.meshgrid(x,y)
    Z      = X**2+Y**2
    h      = 1./(1j*wavelength*distance)*np.exp(1j*k*(distance+Z/2/distance))
    h      = np.fft.fft2(np.fft.fftshift(h))*dx**2
    U1     = np.fft.fft2(np.fft.fftshift(field))
    U2     = h*U1
    result = np.fft.ifftshift(np.fft.ifft2(U2))
    print('hello')
    return result

def angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate angular spectrum based beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    nv,nu  = field.shape
    x      = np.linspace(-nu/2*dx,nu/2*dx,nu)
    y      = np.linspace(-nv/2*dx,nv/2*dx,nv)
    X,Y    = np.meshgrid(x,y)
    Z      = X**2+Y**2
    h      = 1./(1j*wavelength*distance)*np.exp(1j*k*(distance+Z/2/distance))
    h      = np.fft.fft2(np.fft.fftshift(h))*dx**2
    U1     = np.fft.fft2(np.fft.fftshift(field))
    U2     = h*U1
    result = np.fft.ifftshift(np.fft.ifft2(U2))
    return result

def impulse_response_fresnel(field,k,distance,dx,wavelength):
    """
    A definition to calculate impulse response based Fresnel approximation for beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).

    """
    nv,nu  = field.shape
    x      = np.linspace(-nu/2*dx,nu/2*dx,nu)
    y      = np.linspace(-nv/2*dx,nv/2*dx,nv)
    X,Y    = np.meshgrid(x,y)
    Z      = X**2+Y**2
    h      = np.exp(1j*k*distance)/(1j*wavelength*distance)*np.exp(1j*k/2/distance*Z)
    h      = np.fft.fft2(np.fft.fftshift(h))*dx**2
    U1     = np.fft.fft2(np.fft.fftshift(field))
    U2     = h*U1
    result = np.fft.ifftshift(np.fft.ifft2(U2))
    return result

def transfer_function_fresnel(field,k,distance,dx,wavelength):
    """
    A definition to calculate convolution based Fresnel approximation for beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    -----------
    result           : np.complex
                       Final complex field (MxN).

    """
    nv,nu  = field.shape
    fx     = np.linspace(-1./2./dx,1./2./dx,nu)
    fy     = np.linspace(-1./2./dx,1./2./dx,nv)
    FX,FY  = np.meshgrid(fx,fy)
    H      = np.exp(1j*k*distance*(1-(FX*wavelength)**2-(FY*wavelength)**2)**0.5)
    U1     = np.fft.fftshift(np.fft.fft2(np.fft.fftshift(field)))
    U2     = H*U1
    result = np.fft.ifftshift(np.fft.ifft2(np.fft.ifftshift(U2)))
    return result

def band_extended_angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate bandextended angular spectrum based beam propagation. For more Zhang, Wenhui, Hao Zhang, and Guofan Jin. "Band-extended angular spectrum method for accurate diffraction calculation in a wide propagation range." Optics Letters 45.6 (2020): 1543-1546.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    iflag = -1
    eps   = 10**(-12)
    nv,nu = field.shape
    l     = nu*dx
    x     = np.linspace(-l/2,l/2,nu)
    y     = np.linspace(-l/2,l/2,nv)
    X,Y   = np.meshgrid(x,y)
    Z     = X**2+Y**2
    fx    = np.linspace(-1./2./dx,1./2./dx,nu)
    fy    = np.linspace(-1./2./dx,1./2./dx,nv)
    FX,FY = np.meshgrid(fx,fy)
    K     = nu/2/np.amax(fx)
    fcn   = 1./2*(nu/wavelength/np.abs(distance))**0.5
    ss    = np.abs(fcn)/np.amax(np.abs(fx))
    zc    = nu*dx**2/wavelength
    if np.abs(distance) < zc:
        fxn = fx
        fyn = fy
    else:
        fxn = fx*ss
        fyn = fy*ss
    FXN,FYN   = np.meshgrid(fxn,fyn)
    Hn        = np.exp(1j*k*distance*(1-(FXN*wavelength)**2-(FYN*wavelength)**2)**0.5)
    X         = X/np.amax(X)*np.pi
    Y         = Y/np.amax(Y)*np.pi
    t_asmNUFT = nufft2(field,X*ss,Y*ss,sign=iflag,eps=eps)
    result    = nuifft2(Hn*t_asmNUFT,X*ss,Y*ss,sign=-iflag,eps=eps)
    return result

def rayleigh_sommerfeld(field,k,distance,dx,wavelength):
    """
    Definition to compute beam propagation using Rayleigh-Sommerfeld's diffraction formula (Huygens-Fresnel Principle). For more see Section 3.5.2 in Goodman, Joseph W. Introduction to Fourier optics. Roberts and Company Publishers, 2005.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    nv,nu     = field.shape
    x         = np.linspace(-nv*dx/2,nv*dx/2,nv)
    y         = np.linspace(-nu*dx/2,nu*dx/2,nu)
    X,Y       = np.meshgrid(x,y)
    Z         = X**2+Y**2
    result    = np.zeros(field.shape,dtype=np.complex64)
    direction = int(distance/np.abs(distance))
    for i in range(nu):
        for j in range(nv):
            if field[i,j] != 0:
                r01      = np.sqrt(distance**2+(X-X[i,j])**2+(Y-Y[i,j])**2)*direction
                cosnr01  = np.cos(distance/r01)
                result  += field[i,j]*np.exp(1j*k*r01)/r01*cosnr01
    result *= 1./(1j*wavelength)
    return result

def gerchberg_saxton(field,n_iterations,distance,dx,wavelength,slm_range=6.28,propagation_type='IR Fresnel',initial_phase=None):
    """
    Definition to compute a hologram using an iterative method called Gerchberg-Saxton phase retrieval algorithm. For more on the method, see: Gerchberg, Ralph W. "A practical algorithm for the determination of phase from image and diffraction plane pictures." Optik 35 (1972): 237-246.

    Parameters
    ----------
    field            : np.complex64
                       Complex field (MxN).
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.
    slm_range        : float
                       Typically this is equal to two pi. See odak.wave.adjust_phase_only_slm_range() for more.
    propagation_type : str
                       Type of the propagation (IR Fresnel, TR Fresnel, Fraunhofer).
    initial_phase    : np.complex64
                       Phase to be added to the initial value.

    Result
    ---------
    hologram         : np.complex
                       Calculated complex hologram.
    reconstruction   : np.complex
                       Calculated reconstruction using calculated hologram. 
    """
    k              = wavenumber(wavelength)
    target         = calculate_amplitude(field)
    hologram       = generate_complex_field(np.ones(field.shape),0)
    hologram       = zero_pad(hologram)
    if type(initial_phase) == type(None):
        hologram = add_random_phase(hologram)
    else:
        initial_phase = zero_pad(initial_phase) 
        hologram      = add_phase(hologram,initial_phase)
    center         = [int(hologram.shape[0]/2.),int(hologram.shape[1]/2.)]
    orig_shape     = [int(field.shape[0]/2.),int(field.shape[1]/2.)]
    for i in tqdm(range(n_iterations),leave=False):
        reconstruction  = propagate_beam(hologram,k,distance,dx,wavelength,propagation_type)
        new_target      = calculate_amplitude(reconstruction)
        new_target[
                   center[0]-orig_shape[0]:center[0]+orig_shape[0],
                   center[1]-orig_shape[1]:center[1]+orig_shape[1]
                  ]     = target
        reconstruction  = generate_complex_field(new_target,calculate_phase(reconstruction))
        hologram        = propagate_beam(reconstruction,k,-distance,dx,wavelength,propagation_type)
        hologram        = generate_complex_field(1,calculate_phase(hologram)) 
        hologram        = hologram[
                                   center[0]-orig_shape[0]:center[0]+orig_shape[0],
                                   center[1]-orig_shape[1]:center[1]+orig_shape[1],
                                  ]
        hologram        = zero_pad(hologram)
    reconstruction = propagate_beam(hologram,k,distance,dx,wavelength,propagation_type)
    hologram       = hologram[
                              center[0]-orig_shape[0]:center[0]+orig_shape[0],
                              center[1]-orig_shape[1]:center[1]+orig_shape[1]
                             ]
    reconstruction = reconstruction[
                                    center[0]-orig_shape[0]:center[0]+orig_shape[0],
                                    center[1]-orig_shape[1]:center[1]+orig_shape[1]
                                   ]
    return hologram,reconstruction

def point_wise(field,distances,k,dx,wavelength,lens_method='ideal',propagation_method='Bandlimited Angular Spectrum',n_iteration=3):
    """
    Point-wise hologram calculation method. For more Maimone, Andrew, Andreas Georgiou, and Joel S. Kollin. "Holographic near-eye displays for virtual and augmented reality." ACM Transactions on Graphics (TOG) 36.4 (2017): 1-16.

    Parameters
    ----------
    field            : ndarray
                       Complex input field to be converted into a hologram.
    distances        : ndarray
                       Depth map of the input field.
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    dx               : float
                       Pixel pitch.
    wavelength       : float
                       Wavelength of the light.
    lens_model       : str
                       Method to calculate the lens patterns.
    propagation_mode : str
                       Beam propagation method to be used if the lens_model is not equal to `ideal`.
    n_iteration      : int
                       Number of iterations.

    Returns
    ----------
    hologram   : ndarray
                 Generated complex hologram.
    """
    hologram      = np.zeros(field.shape,dtype=np.complex64)
    nx,ny         = field.shape
    cx            = int(nx/2)
    cy            = int(ny/2)
    non_zeros     = np.asarray((np.abs(field)>0).nonzero())
    unique_dist   = np.unique(distances)
    unique_dist   = unique_dist[unique_dist!=0]
    target        = np.zeros((nx,ny),dtype=np.complex64)
    target[cx,cy] = 1.
    lenses        = []
    for distance in tqdm(unique_dist,leave=False):
        if lens_method == 'ideal':
            new_lens = quadratic_phase_function(nx,ny,k,focal=distance,dx=dx)
            lenses.append(new_lens)
        elif lens_method == 'Gerchberg-Saxton':
            new_lens,_ = gerchberg_saxton(
                                          target,
                                          n_iteration,
                                          distance,
                                          dx,
                                          wavelength,
                                          np.pi*2,
                                          propagation_method
                                         )
            lenses.append(new_lens)
    for m in tqdm(range(non_zeros.shape[1]),leave=False):
        i         = int(non_zeros[0,m])
        j         = int(non_zeros[1,m])
        lens_id   = int(np.argwhere(unique_dist==distances[i,j]))
        lens      = lenses[lens_id]
        lens      = np.roll(lens,i-cx,axis=0)
        lens      = np.roll(lens,j-cy,axis=1)
        hologram += lens*field[i,j]
    return hologram

def gerchberg_saxton_3d(fields,n_iterations,distances,dx,wavelength,slm_range=6.28,propagation_type='IR Fresnel',initial_phase=None,target_type='no constraint',coefficients=None):
    """
    Definition to compute a multi plane hologram using an iterative method called Gerchberg-Saxton phase retrieval algorithm. For more on the method, see: Zhou, Pengcheng, et al. "30.4: Multi‐plane holographic display with a uniform 3D Gerchberg‐Saxton algorithm." SID Symposium Digest of Technical Papers. Vol. 46. No. 1. 2015.

    Parameters
    ----------
    fields           : np.complex64
                       Complex fields (MxN).
    distances        : list
                       Propagation distances.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.
    slm_range        : float
                       Typically this is equal to two pi. See odak.wave.adjust_phase_only_slm_range() for more.
    propagation_type : str
                       Type of the propagation (IR Fresnel, TR Fresnel, Fraunhofer).
    initial_phase    : np.complex64
                       Phase to be added to the initial value.
    target_type      : str
                       Target type. `No constraint` targets the input target as is. `Double constraint` follows the idea in this paper, which claims to suppress speckle: Chang, Chenliang, et al. "Speckle-suppressed phase-only holographic three-dimensional display based on double-constraint Gerchberg–Saxton algorithm." Applied optics 54.23 (2015): 6994-7001. 

    Result
    ---------
    hologram         : np.complex
                       Calculated complex hologram.
    """
    k              = wavenumber(wavelength)
    targets        = calculate_amplitude(np.asarray(fields)).astype(np.float)
    hologram       = generate_complex_field(np.ones(targets[0].shape),0)
    hologram       = zero_pad(hologram)
    if type(initial_phase) == type(None):
        hologram = add_random_phase(hologram)
    else:
        initial_phase = zero_pad(initial_phase)
        hologram      = add_phase(hologram,initial_phase)
    center         = [int(hologram.shape[0]/2.),int(hologram.shape[1]/2.)]
    orig_shape     = [int(fields[0].shape[0]/2.),int(fields[0].shape[1]/2.)]
    holograms      = np.zeros((len(distances),hologram.shape[0],hologram.shape[1]),dtype=np.complex64)
    for i in tqdm(range(n_iterations),leave=False):
        for distance_id in tqdm(range(len(distances)),leave=False):
            distance        = distances[distance_id]
            reconstruction  = propagate_beam(hologram,k,distance,dx,wavelength,propagation_type)
            if target_type == 'double constraint':
                if type(coefficients) == type(None):
                    raise Exception("Provide coeeficients of alpha,beta and gamma for double constraint.")
                alpha = coefficients[0]; beta=coefficients[1];gamma=coefficients[2]
                target_current                    = 2*alpha*np.copy(targets[distance_id])-beta*calculate_amplitude(reconstruction)
                target_current[target_current==0] = gamma*np.abs(reconstruction[target_current==0])
            elif target_type == 'no constraint':
                target_current = np.abs(targets[distance_id])
            new_target             = calculate_amplitude(reconstruction)
            new_target[
                       center[0]-orig_shape[0]:center[0]+orig_shape[0],
                       center[1]-orig_shape[1]:center[1]+orig_shape[1]
                      ]            = target_current
            reconstruction         = generate_complex_field(new_target,calculate_phase(reconstruction))
            hologram_layer         = propagate_beam(reconstruction,k,-distance,dx,wavelength,propagation_type)
            hologram_layer         = generate_complex_field(1.,calculate_phase(hologram_layer)) 
            hologram_layer         = hologram_layer[
                                                    center[0]-orig_shape[0]:center[0]+orig_shape[0],
                                                    center[1]-orig_shape[1]:center[1]+orig_shape[1]
                                                   ]
            hologram_layer         = zero_pad(hologram_layer)
            holograms[distance_id] = hologram_layer
        hologram = np.sum(holograms,axis=0)
    hologram = hologram[
                        center[0]-orig_shape[0]:center[0]+orig_shape[0],
                        center[1]-orig_shape[1]:center[1]+orig_shape[1]
                       ]
    return hologram





import math
from odak import np

def rotmatx(angle):
    """
    Definition to generate a rotation matrix along X axis.

    Parameters
    ----------
    angles       : list
                   Rotation angles in degrees.

    Returns
    ----------
    rotx         : ndarray
                    Rotation matrix along X axis.
    """
    angle = np.float(angle)
    angle = np.radians(angle)
    rotx  = np.array([
                      [1.,               0.,               0.],
                      [0.,  math.cos(angle), -math.sin(angle)],
                      [0.,  math.sin(angle),  math.cos(angle)]
                     ],dtype=np.float)
    return rotx

def rotmaty(angle):
    """
    Definition to generate a rotation matrix along Y axis.

    Parameters
    ----------
    angles       : list
                   Rotation angles in degrees.

    Returns
    ----------
    roty         : ndarray
                   Rotation matrix along Y axis.
    """
    angle = np.radians(angle)
    roty  = np.array([
                      [math.cos(angle),  0., math.sin(angle)],
                      [0.,               1.,              0.],
                      [-math.sin(angle), 0., math.cos(angle)]
                     ],dtype=np.float)
    return roty

def rotmatz(angle):
    """
    Definition to generate a rotation matrix along Z axis.

    Parameters
    ----------
    angles       : list
                   Rotation angles in degrees.

    Returns
    ----------
    rotz         : ndarray
                   Rotation matrix along Z axis.
    """
    angle = np.radians(angle)
    rotz  = np.array([
                      [ math.cos(angle), -math.sin(angle), 0.],
                      [ math.sin(angle),  math.cos(angle), 0.],
                      [              0.,               0., 1.]
                     ],dtype=np.float)
    return rotz

def rotate_point(point,angles=[0,0,0],mode='XYZ',origin=[0,0,0],offset=[0,0,0]):
    """
    Definition to rotate a given point. Note that rotation is always with respect to 0,0,0.

    Parameters
    ----------
    point        : ndarray
                   A point.
    angles       : list
                   Rotation angles in degrees. 
    mode         : str
                   Rotation mode determines ordering of the rotations at each axis. There are XYZ,YXZ,ZXY and ZYX modes.
    origin       : list
                   Reference point for a rotation.
    offset       : list
                   Shift with the given offset.

    Returns
    ----------
    result       : ndarray
                   Result of the rotation
    rotx         : ndarray
                   Rotation matrix along X axis.
    roty         : ndarray
                   Rotation matrix along Y axis.
    rotz         : ndarray
                   Rotation matrix along Z axis.
    """
    point  = np.asarray(point)
    point -= np.asarray(origin)
    rotx   = rotmatx(angles[0])
    roty   = rotmaty(angles[1])
    rotz   = rotmatz(angles[2])
    if mode == 'XYZ':
        result = np.dot(rotz,np.dot(roty,np.dot(rotx,point)))
    elif mode == 'XZY':
        result = np.dot(roty,np.dot(rotz,np.dot(rotx,point)))
    elif mode == 'YXZ':
        result = np.dot(rotz,np.dot(rotx,np.dot(roty,point)))
    elif mode == 'ZXY':
        result = np.dot(roty,np.dot(rotx,np.dot(rotz,point)))
    elif mode == 'ZYX':
        result = np.dot(rotx,np.dot(roty,np.dot(rotz,point)))
    result += np.asarray(origin)
    result += np.asarray(offset)
    return result,rotx,roty,rotz

def rotate_points(points,angles=[0,0,0],mode='XYZ',origin=[0,0,0],offset=[0,0,0]):
    """
    Definition to rotate points.

    Parameters
    ----------
    points       : ndarray
                   Points.
    angles       : list
                   Rotation angles in degrees. 
    mode         : str
                   Rotation mode determines ordering of the rotations at each axis. There are XYZ,YXZ,ZXY and ZYX modes.
    origin       : list
                   Reference point for a rotation.
    offset       : list
                   Shift with the given offset.

    Returns
    ----------
    result       : ndarray
                   Result of the rotation   
    """
    points  = np.asarray(points)
    if angles[0] == 0 and angles[1] == 0 and angles[2] ==0:
        result = np.array(offset) + points
        return result
    points -= np.array(origin)
    rotx    = rotmatx(angles[0])
    roty    = rotmaty(angles[1])
    rotz    = rotmatz(angles[2])
    if mode == 'XYZ':
        result = np.dot(rotz,np.dot(roty,np.dot(rotx,points.T))).T
    elif mode == 'XZY':
        result = np.dot(roty,np.dot(rotz,np.dot(rotx,points.T))).T
    elif mode == 'YXZ':
        result = np.dot(rotz,np.dot(rotx,np.dot(roty,points.T))).T
    elif mode == 'ZXY':
        result = np.dot(roty,np.dot(rotx,np.dot(rotz,points.T))).T
    elif mode == 'ZYX':
        result = np.dot(rotx,np.dot(roty,np.dot(rotz,points.T))).T
    result += np.array(origin)
    result += np.array(offset)
    return result

def tilt_towards(location,lookat):
    """
    Definition to tilt surface normal of a plane towards a point.

    Parameters
    ----------
    location     : list
                   Center of the plane to be tilted.
    lookat       : list
                   Tilt towards this point.

    Returns
    ----------
    angles       : list
                   Rotation angles in degrees.
    """
    dx     = location[0]-lookat[0]
    dy     = location[1]-lookat[1]
    dz     = location[2]-lookat[2]
    dist   = np.sqrt(dx**2+dy**2+dz**2)
    phi    = np.arctan2(dy,dx)
    theta  = np.arccos(dz/dist)
    angles = [
               0,
               np.degrees(theta).tolist(),
               np.degrees(phi).tolist()
             ]
    return angles




import math
import bpy
import mathutils
from mathutils import Vector
import numpy as np

def set_rotation(input_obj,angle):
    input_obj.rotation_euler[0] = np.radians(angle[0])
    input_obj.rotation_euler[1] = np.radians(angle[1])
    input_obj.rotation_euler[2] = np.radians(angle[2])
    return input_obj

def set_location(input_obj,location):
    input_obj.location[0] = location[0]
    input_obj.location[1] = location[1]
    input_obj.location[2] = location[2]
    return input_obj

def import_ply(filename,location=[0,0,0],angle=[0,0,0],scale=[1,1,1]):
    """
    Definition to import a PLY object.

    Parameters
    ----------
    filename       : str
                     Path of the PLY file to be imported.
    location       : list
                     location of the imported object along X,Y, and Z axes.
    angle          : list
                     Euler angles for rotating the imported object along X,Y, and Z axes.
    scale          : list
                     Scale of the imported object.

    Returns
    ----------
    ply_object     : blender object
                     Created blender object.
    """
    bpy.ops.import_mesh.ply(filepath=filename)
    ply_object             = bpy.context.view_layer.objects.active
    ply_object.scale[0]    = scale[0]
    ply_object.scale[1]    = scale[1]
    ply_object.scale[2]    = scale[2]
    ply_object.location[0] = location[0]
    ply_object.location[1] = location[1]
    ply_object.location[2] = location[2]
#    ply_object                   = set_location(ply_object,location)
#    ply_object                   = set_rotation(ply_object,angle)
    print('Imported name: ', ply_object.name)
    return ply_object

def intersect_a_ray(p0,p1,ob):
    """
    Definition to create a ray between two points.

    Parameters
    ----------
    p0             : list
                     Starting point of a ray.
    p1             : list
                     Ending point of a ray.
    ob             : blender object
                     Object to be tested for intersection with a ray.

    Returns
    ----------
    result         : list
                     Result of ray intersection test containing state, coalision ray, surface normal.
    loc            : bool/float
                     Location of the intersection. If ray isn't intersecting returns False.
    dist           : float
                     Distance from the starting point of the ray to the intersection point.
    """
    mwi             = ob.matrix_world.inverted()
    ray_begin       = mwi @ Vector((p0[0],p0[1],p0[2]))
    ray_end         = mwi @ Vector((p1[0],p1[1],p1[2]))
    ray_direction   = (ray_end-ray_begin).normalized()
    result          = ob.ray_cast(origin=ray_begin,direction=ray_direction)
    loc             = False
    dist            = 0
    if result[0] == True:
        mw   = ob.matrix_world
        loc  = mw @ result[1]
        dist = ((loc[0]-p0[0])**2+(loc[1]-p0[1])**2+(loc[2]-p0[2])**2)**0.5
    return result,loc,dist

def create_ray_from_two_points(x0y0z0,x1y1z1):
    # Because of Python 2 -> Python 3.
    x0,y0,z0  = x0y0z0
    x1,y1,z1  = x1y1z1
    # Create a vector from two given points.
    point     = np.array([x0,y0,z0],dtype=np.float)
    # Distance between two points.
    s         = np.sqrt( pow((x0-x1),2) + pow((y0-y1),2) + pow((z0-z1),2) )
    if s != 0:
        alpha = (x1-x0)/s
        beta  = (y1-y0)/s
        gamma = (z1-z0)/s
    elif s == 0:
        alpha = float('nan')
        beta  = float('nan')
        gamma = float('nan')
    # Cosines vector
    cosines   = np.array([alpha,beta,gamma],dtype=np.float)
    # Returns vector and the distance.
    ray       = np.array([point,cosines],dtype=np.float)
    return ray,s

def reflect_ray(input_ray,normal,intersection):
    mu               = 1
    div              = pow(normal[0],2)  + pow(normal[1],2) + pow(normal[2],2)
    a                = mu* ( input_ray[1][0]*normal[0]
                           + input_ray[1][1]*normal[1]
                           + input_ray[1][2]*normal[2]) / div
    output_ray       = input_ray.copy()
    output_ray[0][0] = intersection[0]
    output_ray[0][1] = intersection[1]
    output_ray[0][2] = intersection[2]
    output_ray[1][0] = input_ray[1][0] - 2*a*normal[0]
    output_ray[1][1] = input_ray[1][1] - 2*a*normal[1]
    output_ray[1][2] = input_ray[1][2] - 2*a*normal[2]
    p0               = output_ray[0]
    p1               = [
                        output_ray[0][0]+10*output_ray[1][0],
                        output_ray[0][1]+10*output_ray[1][1],
                        output_ray[0][2]+10*output_ray[1][2],
                       ]
    return p0,p1

def quit():
    """
    Definition to quit blender.
    """
    bpy.ops.wm.quit_blender()

def render(fn,exit=False):
    """
    Definition to render a scene, and save it as a PNG file.

    Parameters
    ----------
    fn             : str
                     Filename.
    exit           : bool
                     When set to True blender exits upon rendering completion.
    """
    bpy.context.scene.render.image_settings.file_format   = 'PNG'
    bpy.context.scene.render.filepath                     = fn
    print(fn)
    bpy.data.scenes["Scene"].render.resolution_percentage = 100
    print('Device: %s' % bpy.data.scenes["Scene"].cycles.device)
    bpy.ops.render.render(use_viewport = True, write_still=True)
    if exit == True:
        quit()
    return True

def set_render_type(render_type='rgb'):
    """
    Definition to set the render type.

    Parameters
    ----------
    render_type    : str
                     Set it to `rgb` or `depth` to get either color or depth images.
    """
    bpy.context.scene.use_nodes = True
    tree                        = bpy.context.scene.node_tree
    links                       = tree.links
    for node in tree.nodes:
        tree.nodes.remove(node)
    node_composite              = tree.nodes.new('CompositorNodeComposite')
    node_norm                   = tree.nodes.new('CompositorNodeNormalize')
    node_layers                 = tree.nodes.new('CompositorNodeRLayers')
    node_composite.location     = 800,0
    node_norm.location          = 400,-100
    node_layers.location        = 0,0
    link0                       = links.new(node_layers.outputs[2],node_norm.inputs[0])
    if render_type == 'rgb':
        linkr = links.new(node_layers.outputs[0],node_composite.inputs[0])
    elif render_type == 'depth':
        linkr = links.new(node_norm.outputs[0],node_composite.inputs[0])
    return True

def prepare(resolution=[1920,1080],camera_fov=40.0,camera_location=[0.,0.,-15.],camera_lookat=[0.,0.,0.],clip=[0.1,100000000.],device='GPU',intensity=2.,world_color=[0.3,0.3,0.3]):
    """
    Definition to prepare Blender for renderings.
    """
    bpy.context.scene.render.engine                           = 'CYCLES'
    bpy.context.scene.objects["Camera"].data.lens_unit   = 'FOV'
    bpy.context.scene.objects["Camera"].data.angle       = np.radians(camera_fov)
    bpy.context.scene.objects["Camera"].data.clip_start  = clip[0]
    bpy.context.scene.objects["Camera"].data.clip_end    = clip[1]
    bpy.context.scene.objects["Camera"].location[0]      = camera_location[0]
    bpy.context.scene.objects["Camera"].location[1]      = camera_location[1]
    bpy.context.scene.objects["Camera"].location[2]      = camera_location[2]

    dx                                           = camera_location[0]-camera_lookat[0]
    dy                                           = camera_location[1]-camera_lookat[1]
    dz                                           = camera_location[2]-camera_lookat[2]
    dist                                         = math.sqrt(dx**2+dy**2+dz**2)
    phi                                          = math.atan2(dy,dx)
    theta                                        = math.acos(dz/dist)
    camera_angles                                = [
                                                    0,
                                                    np.degrees(theta),
                                                    np.degrees(phi)
                                                   ]
    bpy.context.scene.objects["Camera"].rotation_euler[0] = np.radians(camera_angles[0])
    bpy.context.scene.objects["Camera"].rotation_euler[1] = np.radians(camera_angles[1])
    bpy.context.scene.objects["Camera"].rotation_euler[2] = np.radians(camera_angles[2])
    bpy.data.scenes["Scene"].render.resolution_x = resolution[0]
    bpy.data.scenes["Scene"].render.resolution_y = resolution[1]
    bpy.data.scenes["Scene"].cycles.device       = device
    bpy.data.worlds["World"].use_nodes           = True
    bpy.data.worlds["World"].node_tree.nodes['Background'].inputs[0].default_value[:3] = world_color
    bpy.data.worlds["World"].node_tree.nodes['Background'].inputs[1].default_value     = intensity
    try:
        delete_object('Cube')
        delete_object('Light')
    except:
        pass
    return camera_angles

def delete_object(label):
    """
    Definition to delete an object from the scene.

    Parameters
    ----------
    label          : str
                     String that identifies the object to be deleted.
    """
    bpy.data.objects.remove(bpy.context.scene.objects[label], do_unlink = True)

# Clear all nodes in a mat
def clear_material( material ):
    if material.node_tree:
        material.node_tree.links.clear()
        material.node_tree.nodes.clear()


def assign_color(fn,obj,color=[0.,0.,0.],label='color'):
    mat                  = bpy.data.materials.new(label)
    mat.use_nodes        = True
    clear_material(mat)
    matnodes             = mat.node_tree.nodes
    links                = mat.node_tree.links
    colors               = matnodes.new('ShaderNodeRGB')
    output               = matnodes.new('ShaderNodeOutputMaterial')
    link                 = links.new(colors.outputs['Color'], output.inputs['Surface'])
    obj.data.materials.append(mat)
    colors.outputs['Color'].default_value = (color[0],color[1],color[2],0)


def assign_texture(fn,obj,label):
    mat            = bpy.data.materials.new(label)
    mat.use_nodes  = True
    clear_material(mat)
    matnodes       = mat.node_tree.nodes
    links          = mat.node_tree.links
    diffuse        = matnodes.new('ShaderNodeBsdfDiffuse')
    output         = matnodes.new('ShaderNodeOutputMaterial')
    tex            = matnodes.new('ShaderNodeTexImage')
    bpy.ops.image.open(filepath=fn)
#    bpy.data.images[0].pack(as_png=True)
    bpy.data.images[0].pack()
    tex.image      = bpy.data.images[0]
    link           = links.new(tex.outputs['Color'], diffuse.inputs['Color'])
    link           = links.new(diffuse.outputs['BSDF'], output.inputs['Surface'])
#    disp           = matnodes['Material Output'].inputs['Displacement']
#    mat.node_tree.links.new(disp, tex.outputs['Color'])
    obj.data.materials.append(mat)
    bpy.data.objects[obj.name].select_set(True)
    bpy.ops.object.mode_set(mode='EDIT')
    bpy.ops.uv.unwrap()
    bpy.ops.object.mode_set(mode='OBJECT')
#    obj.material_slots[0].material.node_tree.nodes[“main image”].image = tex

def create_plane(objname,location,size=[1.,1.]):
    """
    Definition to create a plane.

    Parameters
    ----------
    objname        : str
                     Label of the object to be created.
    location       : list
                     Center of the plane in X,Y,Z coordinates.
    size           : list
                     Size of the plane to be created in X and Y axes.

    Returns
    ----------
    plane          : blender object
                     Created blender object.
    """
    ob             = bpy.ops.mesh.primitive_plane_add(
                                                      size=size[0],
                                                      enter_editmode=False,
                                                      location=location
                                                     )
    plane          = bpy.context.render_layer.obkects.active
    plane.name     = objname
    plane.scale[1] = size[1]/size[0]
    return plane

def create_circle(objname,location,radius=1.):
    """
    Definition to create a circle.

    Parameters
    ----------
    objname        : str
                     Label of the object to be created.
    location       : list
                     Center of the plane in X,Y,Z coordinates.
    radius         : float
                     Radius of the circle.

    Returns
    ----------
    circle         : blender object
                     Created blender object.
    """
    ob             = bpy.ops.mesh.primitive_circle_add(
                                                      radius=radius,
                                                      enter_editmode=False,
                                                      location=location,
                                                      fill_type='TRIFAN'
                                                     )
    circle         = bpy.context.render_layer.obkects.active
    circle.name    = objname
    return circle

# Taken from https://blender.stackexchange.com/questions/67517/possible-to-add-a-plane-using-vertices-via-python
def create_plane_from_meshes(objname,location,size=[1.,1.]):
    """
    Definition to create a plane using simple meshes.

    Parameters
    ----------
    objname        : str
                     Label of the object to be created.
    location       : list
                     Center of the plane in X,Y,Z coordinates.
    size           : list
                     Size of the plane to be created in X and Y axes.

    Returns
    ----------
    myobject       : blender object
                     Created blender object.
    """
    # Define arrays for holding data.
    myvertex = []
    myfaces = []
    # Create all Vertices
    # vertex 0
    mypoint  = [(-size[0]/2., -size[1]/2., 0.0)]
    myvertex.extend(mypoint)
    # vertex 1
    mypoint  = [(size[0]/2., -size[1]/2., 0.0)]
    myvertex.extend(mypoint)
    # vertex 2
    mypoint  = [(-size[0]/2., size[1]/2., 0.0)]
    myvertex.extend(mypoint)
    # vertex 3
    mypoint  = [(size[0]/2., size[1]/2., 0.0)]
    myvertex.extend(mypoint)
    # -------------------------------------
    # Create all Faces
    # -------------------------------------
    myface   = [(0, 1, 3, 2)]
    myfaces.extend(myface)
    mymesh   = bpy.data.meshes.new(objname)
    myobject = bpy.data.objects.new(objname, mymesh)
    bpy.context.collection.objects.link(myobject)
    # Generate mesh data
    mymesh.from_pydata(myvertex, [], myfaces)
    # Calculate the edges
    mymesh.update(calc_edges=True)
    # Set Location
    myobject = set_location(
                            myobject,
                            location
                           )
    return myobject

# Taken from https://blender.stackexchange.com/questions/5898/how-can-i-create-a-cylinder-linking-two-points-with-python
def cylinder_between(start_loc,end_loc,r=0.1,objname='cylinder',color=[0.,0.5,0.,0.]):
    """
    Definition to create a cylinder in between two points with a certain radius.

    Parameters
    ----------
    stat_loc       : list
                     List that contains X,Y and Z start location.
    end_loc        : list
                     List that contains X,Y and Z end location.
    r              : float
                     Radius of the cylinder.
    objname        : str
                     Label of the object to be created.
    color          : list
                     Color of the cylinder (RGBA)

    Returns
    ----------
    obj            : blender object
                     Created cylinder.
    """
    x1              = float(start_loc[0])
    y1              = float(start_loc[1])
    z1              = float(start_loc[2])
    x2              = float(end_loc[0])
    y2              = float(end_loc[1])
    z2              = float(end_loc[2])
    dx              = x2 - x1
    dy              = y2 - y1
    dz              = z2 - z1
    dist            = math.sqrt(dx**2 + dy**2 + dz**2)
    obj             = bpy.ops.mesh.primitive_cylinder_add(
                                                          radius = r,
                                                          depth = dist,
                                                          location = (dx/2 + x1, dy/2 + y1, dz/2 + z1)
                                                         )
    cylinder        = bpy.context.view_layer.objects.active
    cylinder.name   = objname
    phi             = math.atan2(dy, dx)
    theta           = math.acos(dz/dist)
    cylinder.rotation_euler[1] = theta
    cylinder.rotation_euler[2] = phi
    mat             = bpy.data.materials.new('diffuse_texture')
    mat.use_nodes   = True
    clear_material(mat)
    matnodes        = mat.node_tree.nodes
    links           = mat.node_tree.links
    diffuse         = matnodes.new('ShaderNodeBsdfDiffuse')
    diffuse.inputs[0].default_value = [color[0],color[1],color[2],color[3]]
    output          = matnodes.new('ShaderNodeOutputMaterial')
    link            = links.new(diffuse.outputs['BSDF'], output.inputs['Surface'])
    cylinder.data.materials.append(mat)
    return cylinder



"""
``odak.raytracing``
===================
Provides necessary definitions for geometric optics. See "General Ray tracing procedure" from G.H. Spencerand M.V.R.K Murty for the theoratical explanation.

"""
# To get sub-modules.
from .ray import *
from .primitives import *
from .boundary import *



"""
``odak.manager``
===================
Provides necessary definition for submitting, running and gathering compute.

"""

import threading
import dispy
from tqdm import tqdm
from odak.tools import shell_command

def start_server():
    """
    Definition to start a dispynode.py.
    """
    cmd = [
           'dispynode.py',
           '--clean',
           '--daemon'
          ]
    shell_command(cmd)
    return True

class agent():
    def __init__(self,compute,cluster=False,depends=[],server=False):
        """
        Class to submit and run task(s), same class gathers results at the end of a compute.
 
        Parameters
        ----------
        compute        : function
                         Compute function to be run.
        cluster        : bool
                         Set it to True to distribute your job across a cluster or multi CPU cores ofa computer. When set to True, make sure that dispynode.py is running, see https://pgiri.github.io/dispy/dispynode.html .Set it to False to run your tasks locally with a single CPU core of a computer.
        depends        : list
                         List of modules, default is an empty list. Use this only when cluster flag is set to True.

        """
        self.cluster = cluster
        self.compute = compute
        self.depends = depends
        self.server  = server
        self.results = []
        self.jobs    = []
        if self.cluster == True:
            self.job_cluster = dispy.JobCluster(self.compute,depends=self.depends)
        if self.server == True:
            self.server_thread = threading.Thread(target=start_server)
            self.server_thread.start()

    def submit(self,args):
        """
        Definition to submitting jobs.

        Parameters
        ----------
        args           : list
                         Variable arguments to pass to the compute.
        """
        if self.cluster == True:
            job = self.job_cluster.submit(*args)
            self.jobs.append(job)
        elif self.cluster == False:
            job = args
            self.jobs.append(job)

    def run(self):
        """
        Definition to run submitted jobs.

        Returns
        -------

        results       : list
                        Returns results from compute with a list.
        """
        print('Progress of the submitted jobs:')
        pbar         = tqdm(total=len(self.jobs))
        self.results = []
        if self.cluster == False:
           for job in self.jobs:
               arguments = job
               result    = self.compute(*arguments)
               self.results.append(result)
               pbar.update(1)
        if self.cluster == True:
           while len(self.jobs) > 0 :
               for job in self.jobs:
                   if job.status == dispy.DispyJob.Finished:
                       result = job.result
                       self.results.append(result)
                       self.jobs.remove(job)
                       pbar.update(1)
           pbar.close()
           self.job_cluster.wait()
           self.job_cluster.print_status()
        return self.results

    def close(self):
        """
        Definition to close the cluster.
        """
        if self.cluster == True:
            self.job_cluster.close()
        return True



"""
``odak.visualize.blender``
==========================
This is your adapter to visualize things on blender.

"""
import odak.tools as tools
import os,time
from subprocess import Popen, PIPE
from .wrapper import *

def init(headless=False,blend_fn=''):
    """
    Definition to start a blender server. This has been tested with a Blender 2.8. Blender server runs on port 8082 on your local machine.

    Parameters
    ----------
    headless      : bool
                    To set blender to headless mode set it to True.
    blend_fn      : str
                    Filename.

    Returns
    ----------
    proc          : subprocess.Popen
                    Generated process. Make sure to kill the process with a `proc.kill()` once you are done with it.
    outs          : str
                    Outputs of the executed command, returns None when check is set to False.
    """
    directory = os.path.dirname(__file__)
    server_fn = '%s/server.py' % directory
    if headless == False:
        cmd = [
               'blender',
               blend_fn,
               '-P',
               server_fn,
              ]
    elif headless == True:
        cmd = [
               'blender',
               blend_fn,
               '-b',
               '-P',
               server_fn,
              ]
    proc,outs,errs = tools.shell_command(cmd,check=False)
    time.sleep(2)
    return proc

def check(proc):
    """
    Definition to check if blender is still alive and wait for completion. Once completed, this definition kills the process.

    Parameters
    ----------
    proc         : subprocess.Popen
                   Process to be monitored, see init() in odak.visualizer.blender to create one.

    """
    while proc.poll() == None:
        pass
    proc.kill()



"""
``odak.learn.wave``
===================
Provides necessary definitions for neural networks and learning algorithms. The definitions are based on torch framework. Provides necessary definitions for merging geometric optics with wave theory and classical approaches in the wave theory as well. See "Introduction to Fourier Optcs" from Joseph Goodman for the theoratical explanation.

"""
from odak import np
import torch
from .classical import *
from .lens import *

def wavenumber(wavelength):
    """
    Definition for calculating the wavenumber of a plane wave.

    Parameters
    ----------
    wavelength   : float
                   Wavelength of a wave in mm.

    Returns
    ----------
    k            : float
                   Wave number for a given wavelength.
    """
    k = 2*np.pi/wavelength
    return k

def calculate_phase(field,deg=False):
    """ 
    Definition to calculate phase of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : torch.cfloat
                   Electric fields or an electric field.
    deg          : bool
                   If set True, the angles will be returned in degrees.

    Returns
    ----------
    phase        : torch.float
                   Phase or phases of electric field(s) in radians.
    """
    phase = field.imag.atan2(field.real)
    if deg == True:
        phase *= 180./3.14
    return phase

def calculate_amplitude(field):
    """ 
    Definition to calculate amplitude of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : torch.cfloat
                   Electric fields or an electric field.

    Returns
    ----------
    amplitude    : torch.float
                   Amplitude or amplitudes of electric field(s).
    """
    amplitude = torch.abs(field)
    return amplitude

def set_amplitude(field,amplitude):
    """
    Definition to keep phase as is and change the amplitude of a given field.

    Parameters
    ----------
    field        : torch.cfloat
                   Complex field.
    amplitude    : torch.cfloat or torch.float
                   Amplitudes.

    Returns
    ----------
    new_field    : torch.cfloat
                   Complex field.
    """
    amplitude = calculate_amplitude(amplitude)
    phase     = calculate_phase(field)
    new_field = amplitude*torch.cos(phase)+1j*amplitude*torch.sin(phase)
    return new_field

def generate_complex_field(amplitude,phase):
    """
    Definition to generate a complex field with a given amplitude and phase.

    Parameters
    ----------
    amplitude         : ndarray or float
                        Amplitude of the field.
    phase             : ndarray or float
                        Phase of the field.

    Returns
    ----------
    field             : ndarray
                        Complex field.
    """
    if type(phase) == 'torch.Tensor':
        phase     = torch.tensor(phase,requires_grad=True)
    elif type(phase) == type(1.):
        phase     = torch.tensor([phase],requires_grad=True)
    if type(amplitude) == 'torch.Tensor':
        amplitude = torch.tensor(amplitude,requires_grad=True)
    elif type(amplitude) == type(1.):
        amplitude = torch.tensor([amplitude],requires_grad=True)
    field     = amplitude*torch.cos(phase)+1j*amplitude*torch.sin(phase)
    return field

def adjust_phase_only_slm_range(native_range,working_wavelength,native_wavelength):
    """
    Definition for calculating the phase range of the Spatial Light Modulator (SLM) for a given wavelength. Here you prove maximum angle as the lower bound is typically zero. If the lower bound isn't zero in angles, you can use this very same definition for calculating lower angular bound as well.

    Parameters
    ----------
    native_range       : float
                         Native range of the phase only SLM in radians (i.e. two pi).
    working_wavelength : float
                         Wavelength of the illumination source or some working wavelength.
    native_wavelength  : float
                         Wavelength which the SLM is designed for.

    Returns
    ==========
    new_range          : float
                         Calculated phase range in radians.
    """
    new_range = native_range/working_wavelength*native_wavelength
    return new_range

def produce_phase_only_slm_pattern(hologram,slm_range,bits=8,default_range=6.28,illumination=None):
    """
    Definition for producing a pattern for a phase only Spatial Light Modulator (SLM) using a given field.

    Parameters
    ==========
    hologram           : torch.cfloat
                         Input holographic field.
    slm_range          : float
                         Range of the phase only SLM in radians for a working wavelength (i.e. two pi). See odak.wave.adjust_phase_only_slm_range() for more.
    filename           : str
                         Optional variable, if provided the patterns will be save to given location.
    bits               : int
                         Quantization bits.
    default_ramge      : float
                         Default range of phase only SLM.
    illumination       : torch.tensor
                         Spatial illumination distribution.

    Returns
    ==========
    pattern            : torch.cfloat
                         Adjusted phase only pattern.
    hologram_digital   : np.int
                         Digital representation of the hologram.
    """
#    hologram_phase    = calculate_phase(hologram) % default_range
    hologram_phase    = calculate_phase(hologram)
    hologram_phase    = hologram_phase % slm_range
    hologram_phase   /= slm_range
    hologram_phase   *= 2**bits
    hologram_digital  = hologram_phase.detach().clone()
    hologram_phase    = torch.ceil(hologram_phase)
    hologram_phase   *= slm_range/2**bits
    if type(illumination) == type(None):
        A = torch.tensor([1.]).to(hologram_phase.device)
    else:
        A = illumination
    return A*torch.cos(hologram_phase)+A*1j*torch.sin(hologram_phase),hologram_digital



import odak.raytracing as raytracer
import odak.catalog as catalog
import sys
from odak.tools.sample import grid_sample,batch_of_rays

def test():
    sample_entry_points = grid_sample(
                                      no=[4,4],
                                      size=[100.,100.],
                                      center=[0.,0.,0.],
                                      angles=[0.,0.,0.]
                                     )
    sample_exit_points  = grid_sample(
                                      no=[4,4],
                                      size=[100.,100.],
                                      center=[0.,0.,100.],
                                      angles=[0.,0.,0.]
                                     )
    rays                = raytracer.create_ray_from_two_points(
                                                               sample_entry_points,
                                                               sample_exit_points
                                                              )
    detector            = catalog.detectors.plane_detector(
                                                           resolution=[5,5],
                                                           shape=[50,50],
                                                           center=[0.,0.,200.],
                                                           angles=[0.,10.,0.]
                                                          )
    normals,distances   = detector.raytrace(rays)
    print(detector.get_field())
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



from odak import np
import torch

def convert_to_torch(a,grad=True):
    """
    A definition to convert Numpy/Cupy arrays to Torch.

    Parameters
    ----------
    a          : ndarray
                 Input Numpy or Cupy array.
    grad       : bool
                 Set if the converted array requires gradient.

    Returns
    ----------
    c          : torch.Tensor
                 Converted array.
    """
    if np.__name__ == 'cupy':
        b = np.asnumpy(a)
    else:
        b = np.copy(a)
    c = torch.from_numpy(b)
    c.requires_grad_(grad)
    return c

def convert_to_numpy(a):
    """
    A definition to convert Torch/Cupy to Numpy.

    Parameters
    ----------
    a          : torch.Tensor or cupy.ndarray
                 Input Cupy or Torch array.

    Returns
    ----------
    b          : numpy.ndarray
                 Converted array.
    """
    if str(type(a)).find('torch') > 0:
        b = a.to('cpu').detach().numpy()
    elif str(type(a)).find('cupy') > 0:
        b = np.asnumpy(a)
    return b



from odak import np

def cross_product(vector1,vector2):
    """
    Definition to cross product two vectors and return the resultant vector. Used method described under: http://en.wikipedia.org/wiki/Cross_product

    Parameters
    ----------
    vector1      : ndarray
                   A vector/ray.
    vector2      : ndarray
                   A vector/ray.

    Returns
    ----------
    ray          : ndarray
                   Array that contains starting points and cosines of a created ray.
    """
    angle  = np.cross(vector1[1].T,vector2[1].T)
    angle  = np.asarray(angle)
    ray    = np.array([vector1[0],angle],dtype=np.float)
    return ray

def same_side(p1,p2,a,b):
    """
    Definition to figure which side a point is on with respect to a line and a point. See http://www.blackpawn.com/texts/pointinpoly/ for more. If p1 and p2 are on the sameside, this definition returns True.

    Parameters
    ----------
    p1          : list
                  Point(s) to check.
    p2          : list
                  This is the point check against.
    a           : list
                  First point that forms the line.
    b           : list
                  Second point that forms the line.
    """
    ba    = np.subtract(b,a)
    p1a   = np.subtract(p1,a)
    p2a   = np.subtract(p2,a)
    cp1   = np.cross(ba,p1a)
    cp2   = np.cross(ba,p2a)
    test  = np.dot(cp1,cp2)
    if len(p1.shape) > 1:
        return test>=0
    if test >= 0:
        return True
    return False

def distance_between_point_clouds(points0,points1):
    """
    A definition to find distance between every point in one cloud to other points in the other point cloud.
    Parameters
    ----------
    points0     : ndarray
                  Mx3 points.
    points1     : ndarray
                  Nx3 points.

    Returns
    ----------
    distances   : ndarray
                  MxN distances.
    """
    c         = points1.reshape((1,points1.shape[0],points1.shape[1]))
    a         = np.repeat(c,points0.shape[0],axis=0)
    b         = points0.reshape((points0.shape[0],1,points0.shape[1]))
    b         = np.repeat(b,a.shape[1],axis=1)
    distances = np.sqrt(np.sum((a-b)**2,axis=2))
    return distances

def distance_between_two_points(point1,point2):
    """
    Definition to calculate distance between two given points.

    Parameters
    ----------
    point1      : list
                  First point in X,Y,Z.
    point2      : list
                  Second point in X,Y,Z.

    Returns
    ----------
    distance    : float
                  Distance in between given two points.
    """
    point1 = np.asarray(point1)
    point2 = np.asarray(point2)
    if len(point1.shape) == 1 and len(point2.shape) == 1:
        distance = np.sqrt(np.sum((point1-point2)**2))
    elif len(point1.shape) == 2 or len(point2.shape) == 2:
        distance = np.sqrt(np.sum((point1-point2)**2,axis=1))
    return distance

def closest_point_to_a_ray(point,ray):
    """
    Definition to calculate the point on a ray that is closest to given point.

    Parameters
    ----------
    point         : list
                    Given point in X,Y,Z.
    ray           : ndarray
                    Given ray.

    Returns
    ---------
    closest_point : ndarray
                    Calculated closest point.
    """
    from odak.raytracing import propagate_a_ray
    if len(ray.shape) == 2:
        ray = ray.reshape((1,2,3))
    p0            = ray[:,0]
    p1            = propagate_a_ray(ray,1.)
    if len(p1.shape) == 2:
        p1 = p1.reshape((1,2,3))
    p1               = p1[:,0]
    p1               = p1.reshape(3)
    p0               = p0.reshape(3)
    point            = point.reshape(3)
    closest_distance = -np.dot((p0-point),(p1-p0))/np.sum((p1-p0)**2)
    closest_point    = propagate_a_ray(ray,closest_distance)[0]
    return closest_point

def point_to_ray_distance(point,ray_point_0,ray_point_1):
    """
    Definition to find point's closest distance to a line represented with two points.

    Parameters
    ----------
    point       : ndarray
                  Point to be tested.
    ray_point_0 : ndarray
                  First point to represent a line.
    ray_point_1 : ndarray
                  Second point to represent a line.

    Returns
    ----------
    distance    : float
                  Calculated distance.
    """
    distance = np.sum(np.cross((point-ray_point_0),(point-ray_point_1))**2)/np.sum((ray_point_1-ray_point_0)**2)
    return distance



import sys

def test():
    from odak import np
    from odak.wave import gerchberg_saxton,adjust_phase_only_slm_range,produce_phase_only_slm_pattern,calculate_amplitude,wavenumber,double_convergence
    from odak.tools import save_image
    wavelength             = 0.000000532
    dx                     = 0.0000064
    distance               = 2.0
    input_field            = np.zeros((500,500),dtype=np.complex64)
    input_field[0::50,:]  += 1
    iteration_number       = 200
    distance_light_slm     = 2.0
    k                      = wavenumber(wavelength)
    initial_phase          = double_convergence(
                                                input_field.shape[0],
                                                input_field.shape[1],
                                                k,
                                                distance+distance_light_slm,
                                                dx
                                               )
    hologram,reconstructed = gerchberg_saxton(
                                              input_field,
                                              iteration_number,
                                              distance,
                                              dx,
                                              wavelength,
                                              np.pi*2,
                                              'Bandlimited Angular Spectrum',
                                              initial_phase=initial_phase
                                             )
    hologram,_              = produce_phase_only_slm_pattern(
                                                            hologram,
                                                            2*np.pi,
                                                            'output_hologram.png'
                                                           )
    amplitude              = calculate_amplitude(reconstructed)
    save_image(
               'output_amplitude.png',
               amplitude,
               cmin=0,
               cmax=np.amax(amplitude)
              )
    assert True==True

if __name__ == '__main__':
    sys.exit(test())



from odak import np
from odak.tools.vector import distance_between_two_points,closest_point_to_a_ray
from odak.raytracing.primitives import is_it_on_triangle,center_of_triangle,sphere_function,cylinder_function
from odak.raytracing.ray import create_ray_from_two_points,propagate_a_ray

def reflect(input_ray,normal):
    """ 
    Definition to reflect an incoming ray from a surface defined by a surface normal. Used method described in G.H. Spencer and M.V.R.K. Murty, "General Ray-Tracing Procedure", 1961.

    Parameters
    ----------
    input_ray    : ndarray
                   A vector/ray (2x3). It can also be a list of rays (nx2x3).
    normal       : ndarray
                   A surface normal (2x3). It also be a list of normals (nx2x3).

    Returns
    ----------
    output_ray   : ndarray
                   Array that contains starting points and cosines of a reflected ray.
    """
    input_ray = np.asarray(input_ray)
    normal    = np.asarray(normal)
    if len(input_ray.shape) == 2:
        input_ray = input_ray.reshape((1,2,3))
    if len(normal.shape) == 2:
        normal    = normal.reshape((1,2,3))
    mu              = 1
    div             = normal[:,1,0]**2  + normal[:,1,1]**2 + normal[:,1,2]**2
    a               = mu* ( input_ray[:,1,0]*normal[:,1,0]
                          + input_ray[:,1,1]*normal[:,1,1]
                          + input_ray[:,1,2]*normal[:,1,2]) / div 
    n               = np.int(np.amax(np.array([normal.shape[0],input_ray.shape[0]])))
    output_ray      = np.zeros((n,2,3))
    output_ray[:,0] = normal[:,0]
    output_ray[:,1] = input_ray[:,1]-2*a*normal[:,1]
    if output_ray.shape[0] == 1:
       output_ray = output_ray.reshape((2,3))
    return output_ray

def intersect_w_surface(ray,points):
    """
    Definition to find intersection point inbetween a surface and a ray. For more see: http://geomalgorithms.com/a06-_intersect-2.html

    Parameters
    ----------
    ray          : ndarray
                   A vector/ray.
    points       : ndarray
                   Set of points in X,Y and Z to define a planar surface.

    Returns
    ----------
    normal       : ndarray
                   Surface normal at the point of intersection.
    distance     : float
                   Distance in between starting point of a ray with it's intersection with a planar surface.
    """
    points               = np.asarray(points)
    normal               = get_triangle_normal(points)
    if len(ray.shape) == 2:
        ray = ray.reshape((1,2,3))
    if len(points) == 2:
        points = points.reshape((1,3,3))
    if len(normal.shape) == 2:
        normal = normal.reshape((1,2,3))
    f                    = normal[:,0]-ray[:,0]
    distance             = np.dot(normal[:,1],f.T)/np.dot(normal[:,1],ray[:,1].T)
    n                    = np.int(np.amax(np.array([ray.shape[0],normal.shape[0]])))
    normal               = np.zeros((n,2,3))
    normal[:,0]          = ray[:,0]+distance.T*ray[:,1]
    distance             = np.abs(distance)
    if normal.shape[0] == 1:
        normal   = normal.reshape((2,3))
        distance = distance.reshape((1))
    if distance.shape[0] == 1 and len(distance.shape) > 1:
        distance = distance.reshape((distance.shape[1]))
    return normal,distance

def get_triangle_normal(triangle,triangle_center=None):
    """
    Definition to calculate surface normal of a triangle.

    Parameters
    ----------
    triangle        : ndarray
                      Set of points in X,Y and Z to define a planar surface (3,3). It can also be list of triangles (mx3x3).
    triangle_center : ndarray
                      Center point of the given triangle. See odak.raytracing.center_of_triangle for more. In many scenarios you can accelerate things by precomputing triangle centers.

    Returns
    ----------
    normal          : ndarray
                      Surface normal at the point of intersection.
    """
    triangle  = np.asarray(triangle)
    if len(triangle.shape) == 2:
        triangle  = triangle.reshape((1,3,3))
    normal    = np.zeros((triangle.shape[0],2,3))
    direction = np.cross(triangle[:,0]-triangle[:,1],triangle[:,2]-triangle[:,1])
    if type(triangle_center) == type(None):
        normal[:,0] = center_of_triangle(triangle)
    else:
        normal[:,0] = triangle_center
    normal[:,1] = direction/np.sum(direction,axis=1)[0]
    if normal.shape[0] == 1:
        normal = normal.reshape((2,3))
    return normal

def intersect_w_circle(ray,circle):
    """
    Definition to find intersection point of a ray with a circle. Returns False for each variable if the ray doesn't intersect with a given circle. Returns distance as zero if there isn't an intersection.

    Parameters
    ----------
    ray          : ndarray
                   A vector/ray.
    circle       : list
                   A list that contains (0) Set of points in X,Y and Z to define plane of a circle, (1) circle center, and (2) circle radius.

    Returns
    ----------
    normal       : ndarray
                   Surface normal at the point of intersection.
    distance     : float
                   Distance in between a starting point of a ray and the intersection point with a given triangle.
    """
    normal,distance    = intersect_w_surface(ray,circle[0])
    if len(normal.shape) == 2:
        normal = normal.reshape((1,2,3))
    distance_to_center                     = distance_between_two_points(normal[:,0],circle[1])
    distance[np.nonzero(distance_to_center>circle[2])] = 0
    if len(ray.shape) == 2:
        normal = normal.reshape((2,3))
    return normal,distance

def intersect_w_triangle(ray,triangle):
    """
    Definition to find intersection point of a ray with a triangle. Returns False for each variable if the ray doesn't intersect with a given triangle.

    Parameters
    ----------
    ray          : ndarray
                   A vector/ray.
    triangle     : ndarray
                   Set of points in X,Y and Z to define a single triangle.

    Returns
    ----------
    normal       : ndarray
                   Surface normal at the point of intersection.
    distance     : float
                   Distance in between a starting point of a ray and the intersection point with a given triangle.
    """
    normal,distance = intersect_w_surface(ray,triangle)
    if is_it_on_triangle(normal[0],triangle[0],triangle[1],triangle[2]) == False:
        return 0,0
    return normal,distance

def get_sphere_normal(point,sphere):
    """
    Definition to get a normal of a point on a given sphere.

    Parameters
    ----------
    point         : ndarray
                    Point on sphere in X,Y,Z.
    sphere        : ndarray
                    Center defined in X,Y,Z and radius.

    Returns
    ----------
    normal_vector : ndarray
                    Normal vector.
    """
    if len(point.shape) == 1:
        point = point.reshape((1,3))
    normal_vector = create_ray_from_two_points(point,sphere[0:3])
    return normal_vector

def get_cylinder_normal(point,cylinder):
    """
    Parameters
    ----------
    point         : ndarray
                    Point on a cylinder defined in X,Y,Z.

    Returns
    ----------
    normal_vector : ndarray
                    Normal vector.
    """
    cylinder_ray  = create_ray_from_two_points(cylinder[0:3],cylinder[4:7])
    closest_point = closest_point_to_a_ray(
                                           point,
                                           cylinder_ray
                                          )
    normal_vector = create_ray_from_two_points(closest_point,point)
    return normal_vector

def intersection_kernel_for_parametric_surfaces(distance,ray,parametric_surface,surface_function):
    """
    Definition for the intersection kernel when dealing with parametric surfaces.

    Parameters
    ----------
    distance           : float
                         Distance.
    ray                : ndarray
                         Ray.
    parametric_surface : ndarray
                         Array that defines a parametric surface.
    surface_function   : ndarray
                         Function to evaluate a point against a parametric surface.

    Returns
    ----------
    point              : ndarray
                         Location in X,Y,Z after propagation.
    error              : float
                         Error.
    """
    new_ray = propagate_a_ray(ray,distance)
    if len(new_ray) == 2:
        new_ray = new_ray.reshape((1,2,3))
    point = new_ray[:,0]
    error = surface_function(point,parametric_surface)
    return error,point

def propagate_parametric_intersection_error(distance,error):
    """
    Definition to propagate the error in parametric intersection to find the next distance to try.

    Parameters
    ----------
    distance     : list
                   List that contains the new and the old distance.
    error        : list
                   List that contains the new and the old error.
    
    Returns
    ----------
    distance     : list
                   New distance.
    error        : list
                   New error.
    """
    new_distance = distance[1]-error[1]*(distance[1]-distance[0])/(error[1]-error[0])
    distance[0]  = distance[1]
    distance[1]  = np.abs(new_distance)
    error[0]     = error[1]
    return distance,error

def intersect_parametric(ray,parametric_surface,surface_function,surface_normal_function,target_error=0.00000001,iter_no_limit=100000):
    """
    Definition to intersect a ray with a parametric surface.

    Parameters
    ----------
    ray                     : ndarray
                              Ray.
    parametric_surface      : ndarray
                              Parameters of the surfaces.
    surface_function        : function
                              Function to evaluate a point against a surface.
    surface_normal_function : function
                              Function to calculate surface normal for a given point on a surface.
    target_error            : float
                              Target error that defines the precision.  
    iter_no_limit           : int
                              Maximum number of iterations.

    Returns
    ----------
    distance                : float
                              Propagation distance.
    normal                  : ndarray
                              Ray that defines a surface normal for the intersection.
    """
    if len(ray.shape) == 2:
        ray = ray.reshape((1,2,3))
    error        = [150,100]
    distance     = [0,0.1]
    iter_no      = 0
    while np.abs(np.max(np.asarray(error[1]))) > target_error:
        error[1],point  = intersection_kernel_for_parametric_surfaces(
                                                                      distance[1],
                                                                      ray,
                                                                      parametric_surface,
                                                                      surface_function
                                                                     )
        distance,error  = propagate_parametric_intersection_error(
                                                                  distance,
                                                                  error
                                                                 )
        iter_no        += 1
        if iter_no > iter_no_limit:
            return False,False
        if np.isnan(np.sum(point)):
            return False,False
    normal = surface_normal_function(
                                     point,
                                     parametric_surface
                                    )
    return distance[1],normal

def intersect_w_cylinder(ray,cylinder):
    """
    Definition to intersect a ray with a cylinder.

    Parameters
    ----------
    ray        : ndarray
                 A ray definition.
    cylinder   : ndarray
                 A cylinder defined with a center in XYZ and radius of curvature.

    Returns
    ----------
    normal     : ndarray
                 A ray defining surface normal at the point of intersection.
    distance   : float
                 Total optical propagation distance.
    """
    distance,normal = intersect_parametric(
                                           ray,
                                           cylinder,
                                           cylinder_function,
                                           get_cylinder_normal
                                          )
    return normal,distance

def intersect_w_sphere(ray,sphere):
    """
    Definition to intersect a ray with a sphere.

    Parameters
    ----------
    ray        : ndarray
                 A ray definition.
    sphere     : ndarray
                 A sphere defined with a center in XYZ and radius of curvature.

    Returns
    ----------
    normal     : ndarray
                 A ray defining surface normal at the point of intersection.
    distance   : float
                 Total optical propagation distance.
    """
    distance,normal = intersect_parametric(
                                           ray,
                                           sphere,
                                           sphere_function,
                                           get_sphere_normal
                                          )
    return normal,distance



"""
``odak.wave``
===================
Provides necessary definitions for merging geometric optics with wave theory and classical approaches in the wave theory as well. See "Introduction to Fourier Optcs" from Joseph Goodman for the theoratical explanation.

"""
# To get sub-modules.
from .vector import *
from .classical import *
from .lens import *
from odak.tools import save_image

def rayleigh_resolution(diameter,focal=None,wavelength=0.0005):
    """
    Definition to calculate rayleigh resolution limit of a lens with a certain focal length and an aperture. Lens is assumed to be focusing a plane wave at a focal distance.

    Parameter
    ---------
    diameter    : float
                  Diameter of a lens.
    focal       : float
                  Focal length of a lens, when focal length is provided, spatial resolution is provided at the focal plane. When focal length isn't provided angular resolution is provided.
    wavelength  : float
                  Wavelength of light.

    Returns
    --------
    resolution  : float
                  Resolvable angular or spatial spot size, see focal in parameters to know what to expect.

    """
    resolution = 1.22*wavelength/diameter
    if type(focal) != type(None):
        resolution *= focal
    return resolution

def calculate_phase(field,deg=False):
    """ 
    Definition to calculate phase of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.
    deg          : bool
                   If set True, the angles will be returned in degrees.

    Returns
    ----------
    phase        : float
                   Phase or phases of electric field(s) in radians.
    """
    phase = np.angle(field)
    if deg == True:
        phase *= 180./np.pi
    return phase

def calculate_amplitude(field):
    """ 
    Definition to calculate amplitude of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.

    Returns
    ----------
    amplitude    : float
                   Amplitude or amplitudes of electric field(s).
    """
    amplitude = np.abs(field)
    return amplitude

def calculate_intensity(field):
    """
    Definition to calculate intensity of a single or multiple given electric field(s).

    Parameters
    ----------
    field        : ndarray.complex or complex
                   Electric fields or an electric field.

    Returns
    ----------
    intensity    : float
                   Intensity or intensities of electric field(s).
    """
    intensity = np.abs(field)**2
    return intensity

def wavenumber(wavelength):
    """
    Definition for calculating the wavenumber of a plane wave.

    Parameters
    ----------
    wavelength   : float
                   Wavelength of a wave in mm.

    Returns
    ----------
    k            : float
                   Wave number for a given wavelength.
    """
    k = 2*np.pi/wavelength
    return k

def rotationspeed(wavelength,c=3*10**11):
    """
    Definition for calculating rotation speed of a wave (w in A*e^(j(wt+phi))).

    Parameters
    ----------
    wavelength   : float
                   Wavelength of a wave in mm.
    c            : float
                   Speed of wave in mm/seconds. Default is the speed of light in the void!

    Returns
    ----------
    w            : float

    """
    f = c*wavelength
    w = 2*np.pi*f
    return w

def add_random_phase(field):
    """
    Definition for adding a random phase to a given complex field.

    Parameters
    ----------
    field        : np.complex64
                   Complex field.

    Returns
    ----------
    new_field    : np.complex64
                   Complex field.
    """
    random_phase = np.pi*np.random.random(field.shape)
    new_field    = add_phase(field,random_phase)
    return new_field

def add_phase(field,new_phase):
    """
    Definition for adding a phase to a given complex field.

    Parameters
    ----------
    field        : np.complex64
                   Complex field.
    new_phase    : np.complex64
                   Complex phase.

    Returns
    ----------
    new_field    : np.complex64
                   Complex field.
    """
    phase        = calculate_phase(field)
    amplitude    = calculate_amplitude(field)
    new_field    = amplitude*np.cos(phase+new_phase)+1j*amplitude*np.sin(phase+new_phase)
    return new_field

def set_amplitude(field,amplitude):
    """
    Definition to keep phase as is and change the amplitude of a given field.

    Parameters
    ----------
    field        : np.complex64
                   Complex field.
    amplitude    : np.array or np.complex64
                   Amplitudes.

    Returns
    ----------
    new_field    : np.complex64
                   Complex field.
    """
    amplitude = calculate_amplitude(amplitude)
    phase     = calculate_phase(field)
    new_field = amplitude*np.cos(phase)+1j*amplitude*np.sin(phase)
    return new_field

def generate_complex_field(amplitude,phase):
    """
    Definition to generate a complex field with a given amplitude and phase.

    Parameters
    ----------
    amplitude         : ndarray
                        Amplitude of the field.
    phase             : ndarray
                        Phase of the field.

    Returns
    ----------
    field             : ndarray
                        Complex field.
    """
    field = amplitude*np.cos(phase)+1j*amplitude*np.sin(phase)
    return field

def adjust_phase_only_slm_range(native_range,working_wavelength,native_wavelength):
    """
    Definition for calculating the phase range of the Spatial Light Modulator (SLM) for a given wavelength. Here you prove maximum angle as the lower bound is typically zero. If the lower bound isn't zero in angles, you can use this very same definition for calculating lower angular bound as well.

    Parameters
    ----------
    native_range       : float
                         Native range of the phase only SLM in radians (i.e. two pi).
    working_wavelength : float
                         Wavelength of the illumination source or some working wavelength.
    native_wavelength  : float
                         Wavelength which the SLM is designed for.

    Returns
    ==========
    new_range          : float
                         Calculated phase range in radians.
    """
    new_range = native_range/working_wavelength*native_wavelength
    return new_range

def produce_phase_only_slm_pattern(hologram,slm_range,filename=None,bits=8,default_range=6.28,illumination=None):
    """
    Definition for producing a pattern for a phase only Spatial Light Modulator (SLM) using a given field.

    Parameters
    ==========
    hologram           : np.complex64
                         Input holographic field.
    slm_range          : float
                         Range of the phase only SLM in radians for a working wavelength (i.e. two pi). See odak.wave.adjust_phase_only_slm_range() for more.
    filename           : str
                         Optional variable, if provided the patterns will be save to given location.
    bits               : int
                         Quantization bits.
    default_range      : float 
                         Default range of phase only SLM.
    illumination       : np.ndarray
                         Spatial illumination distribution.

    Returns
    ==========
    pattern            : np.complex64
                         Adjusted phase only pattern.
    hologram_digital   : np.int
                         Digital representation of the hologram.
    """
    #hologram_phase   = calculate_phase(hologram) % default_range
    hologram_phase   = calculate_phase(hologram)
    hologram_phase   = hologram_phase % slm_range
    hologram_phase  /= slm_range
    hologram_phase  *= 2**bits
    hologram_phase   = hologram_phase.astype(np.int)
    hologram_digital = np.copy(hologram_phase)
    if type(filename) != type(None):
        save_image(
                   filename,
                   hologram_phase,
                   cmin=0,
                   cmax=2**bits
                  )
    hologram_phase   = hologram_phase.astype(np.float)
    hologram_phase  *= slm_range/2**bits
    if type(illumination) == type(None):
        A = 1.
    else:
        A = illumination
    return A*np.cos(hologram_phase)+A*1j*np.sin(hologram_phase),hologram_digital



"""
``odak.learn.wave``
===================
Provides necessary definitions for neural networks and learning algorithms. The definitions are based on torch framework.
"""
import odak.learn.wave
import odak.learn.tools



from odak import np
import os
import odak.catalog
from odak.tools.transformation import rotate_point
from odak.tools.vector import same_side
from odak.tools.file import load_dictionary
from odak.raytracing.primitives import define_sphere,define_circle
from odak.raytracing.boundary import intersect_w_circle,intersect_w_sphere

class plano_convex_lens():
    """
    A class to represent a plano-convex lens. This is generally useful for raytracing and wave calculations.
    """
    def __init__(self,item='LA1024',location=[0.,0.,0.],rotation=[0.,0.,0.],wavelength=0.000532,meduium='air'):
        """
        Class to represent plano-convex lens.

        Parameters
        ----------
        item        : str
                      Plano convex lens label. Check Thorlabs catalog for labels. There are also json files within library.
        location    : list
                      Location in X,Y, and Z.
        rotation    : list
                      Rotation in X,Y, and Z.
        wavelength  : float
                      Wavelength in mm (default).
        medium      : str
                      Medium that the lens is in. Default is air.
        """
        self.item            = item
        self.location        = np.asarray(location)
        self.rotation        = np.asarray(rotation)
        self.path_to_catalog = "{}/data/plano_convex_lenses.json".format(os.path.dirname(odak.catalog.__file__))
        self.settings        = load_dictionary(self.path_to_catalog)[self.item]
        self.set_variables()
        self.define_geometry()
    
    def set_variables(self):
        """
        A definition to set variables for definining plane-convex lens.
        """
        self.type               = self.settings["type"]
        self.units              = self.settings["units"]
        self.diameter           = self.settings["diameter"]
        self.thickness          = self.settings["center thickness"]
        self.radius             = self.settings["radius of curvature"]
        self.coating            = self.settings["coating"]

    def define_geometry(self):
        """
        A definition to define geometry of a plano-convex lens.
        """
        self.center         = np.array([
                                        0.,
                                        0.,
                                        self.radius-self.thickness 
                                       ])
        self.center,_,_,_   = rotate_point(
                                           self.center,
                                           angles=self.rotation,
                                           offset=self.location
                                          )
        
        self.plane_center   = np.array([0.,0.,0.])+self.location
        self.convex_point   = self.plane_center-self.thickness
        self.convex_surface = define_sphere(
                                            self.center,
                                            self.radius
                                           )
        self.plane_surface  = define_circle(
                                            self.plane_center,
                                            self.diameter,
                                            self.rotation
                                           )
    def intersect(self,ray):
        """
        A  definition to find out which surface to intersect a ray(s).

        Parameters
        ----------
        ray          : ndarray
                       Ray(s) to be intersected. 
        """
        convex_normal,convex_distance = intersect_w_sphere(
                                                           ray,
                                                           self.convex_surface
                                                          )
        plane_normal,plane_distance   = intersect_w_circle(
                                                           ray,
                                                           self.plane_surface
                                                          )
        test_normal                   = convex_normal
        if len(test_normal.shape) < 3:
            test_normal = convex_normal[0]
        is_it_in_lens                 = same_side(
                                                  test_normal,
                                                  self.convex_point,
                                                  self.plane_surface[0][1],
                                                  self.plane_surface[0][2]
                                                 )
        surface_normals               = np.array([convex_normal,plane_normal],dtype=np.float)
        surface_distances             = np.array([convex_distance,plane_distance],dtype=np.float)
        which_surface                 = np.amin(surface_distances,axis=0)
        ids                           = np.where(surface_distances==which_surface)
        ids                           = np.asarray(ids)
#        ids[0]                        = ids[0] | is_it_in_lens
        normal                        = surface_normals[ids[0],ids[1]]
        distance                      = surface_distances[ids[0],ids[1]]
        return normal,distance

    def raytrace(self):
        """
        A definition to raytrace input ray(s) with the plano-convex lens.


        Parameters
        ----------
        rays       : ndarray
                     Input ray(s).
        """
#        which_surface
#        refract
        return output_rays



import plotly.graph_objects as go
import sys
from plotly.subplots import make_subplots
from odak import np
from odak.wave import calculate_phase,calculate_amplitude,calculate_intensity

class surfaceshow():
    """
    A class for general purpose surface plotting using plotly.
    """
    def __init__(self,title='plot',shape=[1000,1000],labels=['x','y','z'],types=['linear','linear','linear'],font_size=12,tick_no=[10,10,10],margin=[65,50,65,90],camera=[1.87,0.88,-0.64],colorbarlength=0.75):
        """
        Class for plotting detectors.

        Parameters
        ---------- 
        title          : str
                         Title of the plot.
        shape          : list
                         Resolution of the figure to be generated.
        labels         : list
                         Labels for x and y axes.
        types          : list
                         Types of axes, it can be `linear`, `log`, `data`, or `category`. For more see ploty.layout.scene.xaxis.type.
        font_size      : int
                         Font size.
        tick_no        : list
                         Number of ticks along each axis.
        margin         : list
                         Margins in plotting.
        camera         : list
                         Scene camera location along X,Y,Z axes.
        colorbarlength : float
                         Length of the colorbar.
        """
        self.settings   = {
                           'title'              : title,
                           'color scale'        : 'Portland',
                           'x label'            : labels[0],
                           'y label'            : labels[1],
                           'z label'            : labels[2],
                           'x axis type'        : types[0],
                           'y axis type'        : types[1],
                           'z axis type'        : types[2],
                           'font size'          : font_size,
                           'x axis tick number' : tick_no[0],
                           'y axis tick number' : tick_no[1],
                           'z axis tick number' : tick_no[2],
                           'width'              : shape[0],
                           'height'             : shape[1],
                           'margin'             : margin,
                           'camera'             : camera,
                           'colorbar length'    : colorbarlength,
                          }
        self.fig = make_subplots(
                                 rows=1,
                                 cols=1,
                                 specs=[
                                        [
                                         {"type": "scene"},
                                        ],
                                       ],
                                )

    def save_image(self,filename):
        """
        Definition to save the figure.

        Parameters
        ----------
        filename    : str
                      Filename.
        """
        self.fig.write_image(filename)

    def show(self):
        """
        Definition to show the plot.
        """
        self.fig.update_layout(
                               autosize = False,
                               width    = self.settings['width'],
                               height   = self.settings['height'],
                               scene    = dict(
                                               aspectmode  = 'manual',
                                               aspectratio = dict(x=1.,y=1.,z=1.),
                                               xaxis=dict(title=self.settings['x label'],type=self.settings['x axis type'],nticks=self.settings['x axis tick number']),
                                               yaxis=dict(title=self.settings['y label'],type=self.settings['y axis type'],nticks=self.settings['y axis tick number']),
                                               zaxis=dict(title=self.settings['z label'],type=self.settings['z axis type'],nticks=self.settings['z axis tick number']),
                                              ),
                               title    = self.settings['title'],
                               font     = dict(
                                               size=self.settings['font size'],
                                              ),
                               margin   = dict(
                                               l=self.settings['margin'][0],
                                               r=self.settings['margin'][1],
                                               b=self.settings['margin'][2],
                                               t=self.settings['margin'][3]
                                              ),
                               scene_camera_eye=dict(
                                                     x=self.settings['camera'][0],
                                                     y=self.settings['camera'][1],
                                                     z=self.settings['camera'][2]
                                                    ),
                              )
        self.fig.show()

    def add_surface(self,data_x,data_y,data_z,label='',mode='lines+markers',opacity=1.,contour=False):
        """
        Definition to add data to the plot.

        Parameters
        ----------
        data_x      : ndarray
                      X axis data to be plotted.
        data_y      : ndarray
                      Y axis data to be plotted.
        label       : str
                      Label of the plot.  
        mode        : str
                      Mode for the plot, it can be either lines+markers, lines or markers.
        opacity     : float
                      Opacity of the plot. The value must be between one to zero. Zero is fully trasnparent, while one is opaque.
        """
        if np.__name__ == 'cupy':
            data_x = np.asnumpy(data_x)
            data_y = np.asnumpy(data_y)
            data_z = np.asnumpy(data_z)
        self.fig.add_trace(
                           go.Surface(
                                      x=data_x,
                                      y=data_y,
                                      z=data_z,
                                      surfacecolor=data_z,
                                      colorscale=self.settings['color scale'],
                                      opacity=opacity,
                                      colorbar=dict(len=self.settings['colorbar length']),
                                      contours= {
                                                 'z':{
                                                      'show':contour,
                                                     },
                                                }
                                     ),
                           row=1,
                           col=1,
                          )

class plotshow():
    """
    A class for general purpose 1D plotting using plotly.
    """
    def __init__(self,subplot_titles=['plot'],shape=[1000,1000],font_size=16,labels=['x','y'],margin=[65,50,65,90],colorbarlength=0.75,rows=1,cols=1):
        """
        Class for plotting detectors.

        Parameters
        ----------
        subplot_titles : list
                         Titles of plots.
                         Resolution of the figure to be generated.
        font_size      : int
                         Font size.                       
        labels         : list
                         Labels for x and y axes.
        margin         : list
                         Margins in plotting.
        colorbarlength : float
                         Length of the colorbar.                         
        """
        self.settings   = {
                           'subplot titles'     : subplot_titles,
                           'color scale'        : 'Portland',
                           'font size'          : font_size,                           
                           'x label'            : labels[0],
                           'y label'            : labels[1],
                           'margin'             : margin,
                           'colorbar length'    : colorbarlength,
                           'width'              : shape[0],
                           'height'             : shape[1],
                          }
        specs           = []
        new_row         = []
        for row in range(rows):
            new_col = []
            for col in range(cols):
                new_col.append({"type":"xy"})
            specs.append(new_col)
        self.fig = make_subplots(
                                 rows=rows,
                                 cols=cols,
                                 specs=specs,
                                 subplot_titles=subplot_titles
                                )
    def show(self):
        """
        Definition to show the plot.
        """
        self.fig.layout = go.Layout(
                                    xaxis=dict(title=self.settings['x label']),
                                    yaxis=dict(title=self.settings['y label'])
                                   )      
        self.fig.update_layout(
                               autosize = False,
                               width    = self.settings['width'],
                               height   = self.settings['height'],
                               scene    = dict(
                                               aspectmode  = 'manual',
                                               aspectratio = dict(x=1.,y=1.,z=1.),
                                              ),
                               font     = dict(
                                               size=self.settings['font size'],
                                              ),
                               margin   = dict(
                                               l=self.settings['margin'][0],
                                               r=self.settings['margin'][1],
                                               b=self.settings['margin'][2],
                                               t=self.settings['margin'][3]
                                              ),

                              )
        self.fig.show()

    def save_image(self,filename):
        """
        Definition to save the figure.

        Parameters
        ----------
        filename    : str
                      Filename.
        """
        self.fig.write_image(filename)

    def add_plot(self,data_x,data_y=None,label='',mode='lines+markers',row=1,col=1):
        """
        Definition to add data to the plot.

        Parameters
        ----------
        data_x      : ndarray
                      X axis data to be plotted.
        data_y      : ndarray
                      Y axis data to be plotted.
        label       : str
                      Label of the plot.  
        mode        : str
                      Mode for the plot, it can be either lines+markers, lines or markers.
        """
        if type(data_y) == type(None):
           data_y = np.arange(0,data_x.shape[0])
        if np.__name__ == 'cupy':
            data_x = np.asnumpy(data_x)
            data_y = np.asnumpy(data_y)
        self.fig.add_trace(
                           go.Scatter(
                                      x=data_x,
                                      y=data_y,
                                      mode=mode,
                                      name=label,
                                     ),
                           row=row,
                           col=col
                          )

class detectorshow():
    """
    A class for visualizing detectors using plotly.
    """
    def __init__(self,row_titles=['Field 1'],subplot_titles=['Amplitude','Phase','Intensity'],title='detector',rows=1,cols=1,show_intensity=False,show_amplitude=True,show_phase=True,shape=[1000,1000],margin=[65,50,65,90]):
        """
        Class for plotting detectors.

        Parameters
        ----------
        subplot_titles : list
                         Titles of plots.
        title          : str
                         Title of plots.
        rows           : int
                         Number of rows/fields.
        show_intensity : bool
                         Flag to show intensity.
        show_amplitude : bool
                         Flag to show amplitude.
        show_phase     : bool
                         Flag to show phase.
        """
        m = 0
        if show_intensity == True:
            m +=1
        if show_amplitude == True:
            m +=1
        if show_phase == True:
            m +=1

        self.settings  = {
                          'title'          : title,
                          'subplot titles' : subplot_titles,
                          'row titles'     : row_titles,
                          'color scale'    : 'Portland',
                          'sub column no'  : m,
                          'column number'  : cols*m,
                          'row number'     : rows,
                          'show amplitude' : show_amplitude,
                          'show phase'     : show_phase,
                          'show intensity' : show_intensity,
                          'width'          : shape[0],
                          'height'         : shape[1],
                          'margin'         : margin
                         }
        specs          = []
        for i in range(0,self.settings["row number"]):
            new_row = []
            for j in range(0,self.settings["column number"]):
                new_row.append({"type": "heatmap"})
            specs.append(new_row)
        self.fig = make_subplots(
                                 rows=self.settings["row number"],
                                 cols=self.settings["column number"],
                                 specs=specs,
                                 subplot_titles=self.settings["subplot titles"],
                                 row_titles=self.settings["row titles"]
                                )
    def show(self):
        """
        Definition to show the plot.
        """
        self.fig.update_layout(
                               autosize = True,
                               width    = self.settings['width'],
                               height   = self.settings['height'],
                               scene    = dict(
                                               aspectmode  = 'manual',
                                               aspectratio = dict(x=1.,y=1.,z=1.),
                                              ),
                               margin   = dict(
                                               l=self.settings['margin'][0],
                                               r=self.settings['margin'][1],
                                               b=self.settings['margin'][2],
                                               t=self.settings['margin'][3]
                                              ),
                              )
        self.fig.show()

    def save_image(self,filename):
        """
        Definition to save the figure. Always first show then save.

        Parameters
        ----------
        filename    : str
                      Filename.
        """
        self.fig.write_image(filename)

    def add_field(self,field,row=1,col=1,showscale=False):
        """
        Definition to add a point to the figure.

        Parameters
        ----------
        field          : ndarray
                         Field to be displayed.
        row            : int
                         Row number.
        col            : int
                         Column number.
        showscale      : bool
                         Set True to show color bar.
        """
        amplitude = calculate_amplitude(field)
        phase     = calculate_phase(field,deg=True)
        intensity = calculate_intensity(field)
        if np.__name__ == 'cupy':
            amplitude = np.asnumpy(amplitude)
            phase     = np.asnumpy(phase)
            intensity = np.asnumpy(intensity)
        col = (col-1)*(self.settings["sub column no"])+1
        if self.settings["show amplitude"] == True:
            self.fig.add_trace(
                               go.Heatmap(
                                          z=amplitude,
                                          colorscale=self.settings['color scale'],
                                          showscale = showscale
                                         ),
                               row=row,
                               col=col
                              )
            col += 1

        if self.settings["show phase"] == True:
            self.fig.add_trace(
                               go.Heatmap(
                                          z=phase,
                                          colorscale=self.settings['color scale'],
                                          showscale = showscale
                                         ),
                               row=row,
                               col=col
                              )
            col += 1

        if self.settings["show intensity"] == True:
            self.fig.add_trace(
                               go.Heatmap(
                                          z=intensity,
                                          colorscale=self.settings['color scale'],
                                          showscale = showscale
                                         ),
                               row=row,
                               col=col
                              )
            col += 1

class rayshow():
    """
    A class for visualizing rays using plotly.
    """
    def __init__(self,rows=1,columns=1,subplot_titles=["Ray visualization"],opacity=0.5,line_width=1.,marker_size=1.,color_scale='Inferno'):
        """
        Class for plotting rays.

        Parameters
        ----------
        rows           : int
                         Number of rows.
        columns        : int
                         Number of columns.
        subplot_titles : list
                         Titles of plots.
        opacity        : float
                         Opacity of the markers or lines.
        line_width     : float
                         Line width of the lines.
        marker_size    : float
                         Marker size of the markers.
        color_scale    : str
                         Color scale to be used. 
        """
        self.settings  = {
                          'rows'           : rows,
                          'color scale'    : color_scale,
                          'columns'        : columns,
                          'subplot titles' : subplot_titles,
                          'opacity'        : opacity,
                          'line width'     : line_width,
                          'marker size'    : marker_size
                         }
        specs          = []
        for i in range(0,columns):
            new_row = []
            for j in range(0,rows):
                new_row.append({"type": "scene"},)
            specs.append(new_row)
        self.fig = make_subplots(
                                 rows=self.settings['rows'],
                                 cols=self.settings['columns'],
                                 subplot_titles=self.settings['subplot titles'],
                                 specs=specs
                                )

    def show(self):
        """
        Definition to show the plot.
        """
        self.fig.update_layout(
                               scene = dict(
                                            aspectmode  = 'manual',
                                            aspectratio = dict(x=1.,y=1.,z=1.),
                                           ),
                              )
        self.fig.show()

    def add_point(self,point,row=1,column=1,color='red'):
        """
        Definition to add a point to the figure.

        Parameters
        ----------
        point          : ndarray
                         Point(s).
        row            : int
                         Row number of the figure.
        column         : int
                         Column number of the figure.

        """
        if np.__name__ == 'cupy':
            point = np.asnumpy(point)
        self.fig.add_trace(
                           go.Scatter3d(
                                        x=point[:,0].flatten(),
                                        y=point[:,1].flatten(),
                                        z=point[:,2].flatten(),
                                        mode='markers',
                                        marker=dict(
                                                    size=self.settings["marker size"],
                                                    color=color,
                                                    opacity=self.settings["opacity"]
                                                   ),
                                       ),
                           row=row,
                           col=column
                          )

    def add_line(self,point_start,point_end,row=1,column=1,color='red'):
        """
        Definition to add a ray to the figure.

        Parameters
        ----------
        point_start    : ndarray
                         Starting point(s).
        point_end      : ndarray
                         Ending point(s).
        row            : int
                         Row number of the figure.
        column         : int
                         Column number of the figure.
        color          : str
                         Color of the lune to be drawn.
        """
        if np.__name__ == 'cupy':
            point_start = np.asnumpy(point_start)
            point_end   = np.asnumpy(point_end)
        if len(point_start.shape) == 1:
            point_start = point_start.reshape((1,3))
        if len(point_end.shape) == 1:
            point_end   = point_end.reshape((1,3))
        if point_start.shape != point_end.shape:
            print('Size mismatch in line plot. Sizes are {} and {}.'.format(point_start.shape,point_end.shape))
            sys.exit()
        for point_id in range(0,point_start.shape[0]):
            points = np.array(
                              [
                               point_start[point_id],
                               point_end[point_id]
                              ]
                             )
            points = points.reshape((2,3))
            if np.__name__ == 'cupy':
                points = np.asnumpy(points)
            self.fig.add_trace(
                               go.Scatter3d(
                                            x=points[:,0],
                                            y=points[:,1],
                                            z=points[:,2],
                                            mode='lines',
                                            line=dict(
                                                      width=self.settings["line width"],
                                                      color=color,
                                                     ),
                                            opacity=self.settings["opacity"]
                                           ),
                               row=row,
                               col=column
                              )

    def add_surface(self,data_x,data_y,data_z,surface_color,row=1,column=1,label='',mode='lines+markers',opacity=1.,contour=False):
        """
        Definition to add data to the plot.

        Parameters
        ----------
        data_x        : ndarray
                        X axis data to be plotted.
        data_y        : ndarray
                        Y axis data to be plotted.
        data_z        : ndarray
                        Z axis data to be plotted.
        surface_color : ndarray
                        Colors of the surface.
        label         : str
                        Label of the plot.  
        mode          : str
                        Mode for the plot, it can be either lines+markers, lines or markers.
        opacity       : float
                        Opacity of the plot. The value must be between one to zero. Zero is fully trasnparent, while one is opaque.
        """
        if np.__name__ == 'cupy':
            data_x        = np.asnumpy(data_x)
            data_y        = np.asnumpy(data_y)
            data_z        = np.asnumpy(data_z)
            surface_color = np.asnumpy(surface_color)
        self.fig.add_trace(
                           go.Surface(
                                      x=data_x,
                                      y=data_y,
                                      z=data_z,
                                      surfacecolor=surface_color,
                                      colorscale=self.settings['color scale'],
                                      opacity=opacity,
                                      contours= {
                                                 'z':{
                                                      'show':contour,
                                                     },
                                                }
                                     ),
                           row=row,
                           col=column,
                          )



from odak import np
from odak.tools import rotate_points

def create_ray(x0y0z0,abg):
    """
    Definition to create a ray.

    Parameters
    ----------
    x0y0z0       : list
                   List that contains X,Y and Z start locations of a ray.
    abg          : list
                   List that contaings angles in degrees with respect to the X,Y and Z axes.

    Returns
    ----------
    ray          : ndarray
                   Array that contains starting points and cosines of a created ray.
    """
    # Due to Python 2 -> Python 3.
    x0,y0,z0         = x0y0z0
    alpha,beta,gamma = abg
    # Create a vector with the given points and angles in each direction
    point            = np.array([x0,y0,z0],dtype=np.float)
    alpha            = np.cos(np.radians(alpha))
    beta             = np.cos(np.radians(beta))
    gamma            = np.cos(np.radians(gamma))
    # Cosines vector.
    cosines          = np.array([alpha,beta,gamma],dtype=np.float)
    ray              = np.array([point,cosines],dtype=np.float)
    return ray

def create_ray_from_two_points(x0y0z0,x1y1z1):
    """
    Definition to create a ray from two given points. Note that both inputs must match in shape.

    Parameters
    ----------
    x0y0z0       : list
                   List that contains X,Y and Z start locations of a ray (3). It can also be a list of points as well (mx3). This is the starting point.
    x1y1z1       : list
                   List that contains X,Y and Z ending locations of a ray (3). It can also be a list of points as well (mx3). This is the end point.

    Returns
    ----------
    ray          : ndarray
                   Array that contains starting points and cosines of a created ray.
    """
    x0y0z0 = np.asarray(x0y0z0,dtype=np.float)
    x1y1z1 = np.asarray(x1y1z1,dtype=np.float)
    if len(x0y0z0.shape) == 1:
        x0y0z0 = x0y0z0.reshape((1,3))
    if len(x1y1z1.shape) == 1:
        x1y1z1 = x1y1z1.reshape((1,3))
    xdiff        = x1y1z1[:,0]-x0y0z0[:,0]
    ydiff        = x1y1z1[:,1]-x0y0z0[:,1]
    zdiff        = x1y1z1[:,2]-x0y0z0[:,2]
    s            = np.sqrt(xdiff**2+ydiff**2+zdiff**2)
    s[s==0]      = np.NaN
    cosines      = np.zeros((xdiff.shape[0],3))
    cosines[:,0] = xdiff/s 
    cosines[:,1] = ydiff/s
    cosines[:,2] = zdiff/s
    ray          = np.zeros((xdiff.shape[0],2,3),dtype=np.float)
    ray[:,0]     = x0y0z0
    ray[:,1]     = cosines
    if ray.shape[0] == 1:
        ray = ray.reshape((2,3))
    return ray

def create_ray_from_angles(point,angles,mode='XYZ'):
    """
    Definition to create a ray from a point and angles.

    Parameters
    ----------
    point      : ndarray
                 Point in X,Y and Z.
    angles     : ndarray
                 Angles with X,Y,Z axes in degrees. All zeros point Z axis.
    mode       : str
                 Rotation mode determines ordering of the rotations at each axis. There are XYZ,YXZ    ,ZXY and ZYX modes.

    Returns
    ----------
    ray        : ndarray
                 Created ray.
    """
    if len(point.shape) == 1:
        point  = point.reshape((1,3))
    new_point       = np.zeros(point.shape)
    new_point[:,2] += 5.
    new_point       = rotate_points(new_point,angles,mode=mode,offset=point[:,0])
    ray             = create_ray_from_two_points(point,new_point)
    if ray.shape[0] == 1:
        ray = ray.reshape((2,3))
    return ray

def propagate_a_ray(ray,distance):
    """
    Definition to propagate a ray at a certain given distance.

    Parameters
    ----------
    ray        : ndarray
                 A ray.
    distance   : float
                 Distance.

    Returns
    ----------
    new_ray    : ndarray
                 Propagated ray.
    """
    if len(ray.shape) == 2:
        ray = ray.reshape((1,2,3))
    new_ray        = np.copy(ray)
    new_ray[:,0,0] = distance*new_ray[:,1,0] + new_ray[:,0,0]
    new_ray[:,0,1] = distance*new_ray[:,1,1] + new_ray[:,0,1]
    new_ray[:,0,2] = distance*new_ray[:,1,2] + new_ray[:,0,2]
    if new_ray.shape[0] == 1:
        new_ray = new_ray.reshape((2,3))
    return new_ray

def calculate_intersection_of_two_rays(ray0,ray1):
    """
    Definition to calculate the intersection of two rays.

    Parameters
    ----------
    ray0       : ndarray
                 A ray.
    ray1       : ndarray
                 A ray.

    Returns
    ----------
    point      : ndarray
                 Point in X,Y,Z.
    distances  : ndarray
                 Distances.
    """
    A = np.array([
                  [float(ray0[1][0]),float(ray1[1][0])],
                  [float(ray0[1][1]),float(ray1[1][1])],
                  [float(ray0[1][2]),float(ray1[1][2])]
                 ])
    B = np.array([
                  ray0[0][0]-ray1[0][0],
                  ray0[0][1]-ray1[0][1],
                  ray0[0][2]-ray1[0][2]
                 ])
    distances = np.linalg.lstsq(A,B)[0]
    if np.allclose(np.dot(A,distances),B) == False:
        distances = np.array([0,0])
    distances = distances[np.argsort(-distances)]
    point     = propagate_a_ray(ray0,distances[0])[0]
    return point,distances

def find_nearest_points(ray0,ray1):
    """
    Find the nearest points on given rays with respect to the other ray.

    Parameters
    ----------
    ray0       : ndarray
                 A ray.
    ray1       : ndarray
                 A ray.

    Returns
    ----------
    c0         : ndarray
                 Closest point on ray0.
    c1         : ndarray
                 Closest point on ray1.
    """
    p0 = ray0[0].reshape(3,)
    d0 = ray0[1].reshape(3,)
    p1 = ray1[0].reshape(3,)
    d1 = ray1[1].reshape(3,)
    n  = np.cross(d0, d1)
    if np.all(n)==0:
      point, distances = calculate_intersection_of_two_rays(ray0,ray1)
      c0 = c1 = point
    else:
      n0 = np.cross(d0,n)
      n1 = np.cross(d1,n)
      c0 = p0+(np.dot((p1-p0), n1)/np.dot(d0, n1))*d0
      c1 = p1+(np.dot((p0-p1), n0)/np.dot(d1, n0))*d1
    return c0,c1



from PIL import Image
import odak.tools
import torch
import numpy as np_cpu

def load_image(fn):
    """
    Definition to load an image from a given location as a Numpy array.

    Parameters
    ----------
    fn           : str
                   Filename.

    Returns
    ----------
    image        :  ndarray
                    Image loaded as a Numpy array.

    """
    image = Image.open(fn)
    image = np_cpu.array(image)
    image = torch.from_numpy(image)
    return image

def save_image(fn,img,cmin=0,cmax=255):
    """
    Definition to save a Numpy array as an image.

    Parameters
    ----------
    fn           : str
                   Filename.
    img          : ndarray
                   A numpy array with NxMx3 or NxMx1 shapes.
    cmin         : int
                   Minimum value that will be interpreted as 0 level in the final image.
    cmax         : int
                   Maximum value that will be interpreted as 255 level in the final image.

    Returns
    ----------
    bool         :  bool
                    True if successful.

    """
    img = img.cpu().detach().numpy()
    return odak.tools.save_image(fn,img,cmin,cmax)



from odak import np
import torch, torch.fft
from .__init__ import set_amplitude, produce_phase_only_slm_pattern, generate_complex_field, calculate_amplitude
from odak.wave import wavenumber
from odak.learn.tools import zero_pad, crop_center
from tqdm import tqdm

def propagate_beam(field,k,distance,dx,wavelength,propagation_type='IR Fresnel'):
    """
    Definitions for Fresnel impulse respone (IR), Fresnel Transfer Function (TF), Fraunhofer diffraction in accordence with "Computational Fourier Optics" by David Vuelz.

    Parameters
    ==========
    field            : torch.complex128
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.
    propagation_type : str
                       Type of the propagation (IR Fresnel, TR Fresnel, Fraunhofer).

    Returns
    =======
    result           : torch.complex128
                       Final complex field (MxN).
    """
    if propagation_type == 'IR Fresnel':
        result = impulse_response_fresnel(field,k,distance,dx,wavelength)
    elif propagation_type == 'Bandlimited Angular Spectrum':
        result = band_limited_angular_spectrum(field,k,distance,dx,wavelength)
    elif propagation_type == 'TR Fresnel':
        result = transfer_function_fresnel(field,k,distance,dx,wavelength)
    elif propagation_type == 'Fraunhofer':
        nv, nu = field.shape[-1], field.shape[-2]
        x      = torch.linspace(-nv*dx/2, nv*dx/2, nv, dtype=torch.float32)
        y      = torch.linspace(-nu*dx/2, nu*dx/2, nu, dtype=torch.float32)
        Y, X   = torch.meshgrid(y, x)
        Z      = torch.pow(X,2) + torch.pow(Y,2)
        c      = 1./(1j*wavelength*distance)*torch.exp(1j*k*0.5/distance*Z)
        c      = c.to(field.device)
        result = c*torch.fft.ifftshift(torch.fft.fft2(torch.fft.fftshift(field)))*pow(dx,2)
    return result


def transfer_function_fresnel(field,k,distance,dx,wavelength):
    """
    A definition to calculate convolution based Fresnel approximation for beam propagation.

    Parameters
    ----------
    field            : troch.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    ---------
    result           : torch.complex
                       Final complex field (MxN).

    """
    distance  = torch.tensor([distance]).to(field.device)
    nv, nu    = field.shape[-1], field.shape[-2]
    fx        = torch.linspace(-1./2./dx,1./2./dx,nu,dtype=torch.float32).to(field.device)
    fy        = torch.linspace(-1./2./dx,1./2./dx,nv,dtype=torch.float32).to(field.device)
    FY, FX    = torch.meshgrid(fx, fy)
    H         = torch.exp(1j*k*distance*(1-(FX*wavelength)**2-(FY*wavelength)**2)**0.5)
    H         = H.to(field.device)
    U1        = torch.fft.fftshift(torch.fft.fft2(torch.fft.fftshift(field)))
    U2        = H*U1
    result    = torch.fft.ifftshift(torch.fft.ifft2(torch.fft.ifftshift(U2)))
    return result


def band_limited_angular_spectrum(field,k,distance,dx,wavelength):
    """
    A definition to calculate bandlimited angular spectrum based beam propagation. For more Matsushima, Kyoji, and Tomoyoshi Shimobaba. "Band-limited angular spectrum method for numerical simulation of free-space propagation in far and near fields." Optics express 17.22 (2009): 19662-19673.

    Parameters
    ----------
    field            : torch.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : torch.complex
                       Final complex field (MxN).
    """
    assert True==False,"Refer to Issue 19 for more. This definition is unreliable."
    nv, nu    = field.shape[-1], field.shape[-2]
    x         = torch.linspace(-nv*dx/2, nv*dx/2, nv, dtype=torch.float32)
    y         = torch.linspace(-nu*dx/2, nu*dx/2, nu, dtype=torch.float32)
    Y, X      = torch.meshgrid(y, x)
    Z         = torch.pow(X,2) + torch.pow(Y,2)
    distance  = torch.FloatTensor([distance])
    h         = 1./(1j*wavelength*distance)*torch.exp(1j*k*(distance+Z/2/distance))
    h         = torch.fft.fft2(torch.fft.fftshift(h)) * pow(dx, 2)
    h         = h.to(field.device)
    flimx     = torch.ceil(1/(((2*distance*(1./(nv)))**2+1)**0.5*wavelength))
    flimy     = torch.ceil(1/(((2*distance*(1./(nu)))**2+1)**0.5*wavelength))
    mask      = torch.zeros((nu,nv), dtype=torch.cfloat).to(field.device)
    mask[...] = torch.logical_and(torch.lt(torch.abs(X), flimx), torch.lt(torch.abs(Y), flimy))
    mask      = set_amplitude(h, mask)
    U1        = torch.fft.fft2(torch.fft.fftshift(field))
    U2        = mask * U1
    result    = torch.fft.ifftshift(torch.fft.ifft2(U2))
    return result

def impulse_response_fresnel(field,k,distance,dx,wavelength):
    """
    A definition to calculate impulse response based Fresnel approximation for beam propagation.

    Parameters
    ----------
    field            : np.complex
                       Complex field (MxN).
    k                : odak.wave.wavenumber
                       Wave number of a wave, see odak.wave.wavenumber for more.
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.

    Returns
    =======
    result           : np.complex
                       Final complex field (MxN).
    """
    assert True==False,"Refer to Issue 19 for more. This definition is unreliable."
    nv, nu   = field.shape[-1], field.shape[-2]
    x        = torch.linspace(-nu/2*dx,nu/2*dx,nu)
    y        = torch.linspace(-nv/2*dx,nv/2*dx,nv)
    X,Y      = torch.meshgrid(x,y)
    Z        = X**2+Y**2
    distance = torch.tensor([distance]).to(field.device)
    h        = torch.exp(1j*k*distance)/(1j*wavelength*distance)*torch.exp(1j*k/2/distance*Z)
    h        = torch.fft.fft2(torch.fft.fftshift(h))*dx**2
    h        = h.to(field.device)
    U1       = torch.fft.fft2(torch.fft.fftshift(field))
    U2       = h*U1
    result   = torch.fft.ifftshift(torch.fft.ifft2(U2))
    return result

def gerchberg_saxton(field,n_iterations,distance,dx,wavelength,slm_range=6.28,propagation_type='IR Fresnel'):
    """
    Definition to compute a hologram using an iterative method called Gerchberg-Saxton phase retrieval algorithm. For more on the method, see: Gerchberg, Ralph W. "A practical algorithm for the determination of phase from image and diffraction plane pictures." Optik 35 (1972): 237-246.

    Parameters
    ----------
    field            : torch.cfloat
                       Complex field (MxN).
    distance         : float
                       Propagation distance.
    dx               : float
                       Size of one single pixel in the field grid (in meters).
    wavelength       : float
                       Wavelength of the electric field.
    slm_range        : float
                       Typically this is equal to two pi. See odak.wave.adjust_phase_only_slm_range() for more.
    propagation_type : str
                       Type of the propagation (IR Fresnel, TR Fresnel, Fraunhofer).

    Result
    ---------
    hologram         : torch.cfloat
                       Calculated complex hologram.
    reconstruction   : torch.cfloat
                       Calculated reconstruction using calculated hologram. 
    """
    k              = wavenumber(wavelength)
    reconstruction = field
    for i in range(n_iterations):
        hologram       = propagate_beam(reconstruction,k,-distance,dx,wavelength,propagation_type)
        hologram,_     = produce_phase_only_slm_pattern(hologram,slm_range)
        reconstruction = propagate_beam(hologram,k,distance,dx,wavelength,propagation_type)
        reconstruction = set_amplitude(hologram,field)
    reconstruction = propagate_beam(hologram,k,distance,dx,wavelength,propagation_type)
    return hologram,reconstruction

def stochastic_gradient_descent(field,wavelength,distance,dx,resolution,propogation_type,n_iteration=100,loss_function=None,cuda=False,learning_rate=0.1):
    """
    Definition to generate phase and reconstruction from target image via stochastic gradient descent.

    Parameters
    ----------
    field                   : ndarray
                              Input field as Numpy array.
    wavelength              : double
                              Set if the converted array requires gradient.
    distance                : double
                              Hologaram plane distance wrt SLM plane
    dx                      : float
                              SLM pixel pitch
    resolution              : array
                              SLM resolution
    propogation type        : str
                              Type of the propagation (IR Fresnel, Angular Spectrum, Bandlimited Angular Spectrum, TR Fresnel, Fraunhofer)
    n_iteration:            : int
                              Max iteratation 
    loss_function:          : function
                              If none it is set to be l2 loss
    cuda                    : boolean
                              GPU enabled
    learning_rate           : float
                              Learning rate.

    Returns
    ----------
    hologram                : torch.Tensor
                              Phase only hologram as torch array

    reconstruction_intensity: torch.Tensor
                              Reconstruction as torch array

    """
    torch.cuda.empty_cache()
    torch.manual_seed(0)
    device    = torch.device("cuda" if cuda else "cpu") 
    field     = field.to(device)
    phase     = torch.rand(resolution[0],resolution[1]).detach().to(device).requires_grad_()
    amplitude = torch.ones(resolution[0],resolution[1],requires_grad=False).to(device)
    k         = wavenumber(wavelength)
    optimizer = torch.optim.Adam([{'params': [phase]}],lr=learning_rate)
    if type(loss_function) == type(None):
        loss_function = torch.nn.MSELoss().to(device)
    t = tqdm(range(n_iteration),leave=False)
    for i in t:
        optimizer.zero_grad()
        hologram                 = generate_complex_field(amplitude,phase)
        hologram_padded          = zero_pad(hologram)
        reconstruction_padded    = propagate_beam(hologram_padded,k,distance,dx,wavelength,propogation_type)
        reconstruction           = crop_center(reconstruction_padded)
        reconstruction_intensity = calculate_amplitude(reconstruction)**2
        loss                     = loss_function(reconstruction_intensity,field)
        description              = "Iteration: {} loss:{:.4f}".format(i,loss.item())
        loss.backward(retain_graph=True)
        optimizer.step()
        t.set_description(description)
    print(description)
    torch.no_grad()
    hologram              = generate_complex_field(amplitude,phase)
    hologram_padded       = zero_pad(hologram)
    reconstruction_padded = propagate_beam(hologram_padded,k,distance,dx,wavelength,propogation_type)
    reconstruction        = crop_center(reconstruction_padded)
    hologram              = crop_center(hologram_padded)
    return hologram.detach(), reconstruction.detach()



import subprocess
import os
import json
import pathlib
import scipy.misc
from odak import np
from PIL import Image

def resize_image(img,target_size):
    if np.__name__ == 'cupy':
        import numpy
        img = np.asnumpy(img)
    img = scipy.misc.imresize(img, (target_size[0],target_size[1]))
    if np.__name__ == 'cupy':
        img = np.asarray(img)
    return img

def save_image(fn,img,cmin=0,cmax=255):
    """
    Definition to save a Numpy array as an image.

    Parameters
    ----------
    fn           : str
                   Filename.
    img          : ndarray
                   A numpy array with NxMx3 or NxMx1 shapes.
    cmin         : int
                   Minimum value that will be interpreted as 0 level in the final image.
    cmax         : int
                   Maximum value that will be interpreted as 255 level in the final image.

    Returns
    ----------
    bool         :  bool
                    True if successful.

    """
    input_img                  = np.copy(img).astype(np.float)
    colorflag                  = False
    if  len(input_img.shape) == 3:
        if input_img.shape[2] > 1:
           input_img        = input_img[:,:,0:3]
           colorflag        = True
    input_img[input_img<cmin]  = cmin
    input_img[input_img>cmax]  = cmax
    input_img                 /= cmax
    input_img                 *= 255
    input_img                  = input_img.astype(np.uint8)
    if np.__name__ == 'cupy':
        input_img = np.asnumpy(input_img)
    if colorflag==True:
       result_img = Image.fromarray(input_img)
    elif colorflag == False:
       result_img = Image.fromarray(input_img).convert("L")
    result_img.save(fn)
    return True

def load_image(fn):
    """ 
    Definition to load an image from a given location as a Numpy array.

    Parameters
    ----------
    fn           : str
                   Filename.

    Returns
    ----------
    image        :  ndarray
                    Image loaded as a Numpy array.

    """
    image = Image.open(fn)
    return np.array(image)

def shell_command(cmd,cwd='.',timeout=None,check=True):
    """
    Definition to initiate shell commands.

    Parameters
    ----------
    cmd          : list
                   Command to be executed. 
    cwd          : str
                   Working directory.
    timeout      : int
                   Timeout if the process isn't complete in the given number of seconds.
    check        : bool
                   Set it to True to return the results and to enable timeout.

    Returns
    ----------
    proc         : subprocess.Popen
                   Generated process.
    outs         : str
                   Outputs of the executed command, returns None when check is set to False.
    errs         : str
                   Errors of the executed command, returns None when check is set to False.

    """
    proc = subprocess.Popen(
                            cmd,
                            cwd=cwd,
                            stdout=subprocess.PIPE
                           )
    if check == False:
        return proc,None,None
    try:
        outs, errs = proc.communicate(timeout=timeout)
    except subprocess.TimeoutExpired:
        proc.kill()
        outs, errs = proc.communicate()
    return proc,outs,errs

def check_directory(directory):
    """
    Definition to check if a directory exist. If it doesn't exist, this definition will create one.

    Parameters
    ----------
    directory     : str
                    Full directory path.
    """
    if not os.path.exists(directory):
       os.makedirs(directory)
       return False
    return True

def save_dictionary(settings,filename):
    """
    Definition to load a dictionary (JSON) file.


    Parameters
    ----------
    settings      : dict
                    Dictionary read from the file.
    filename      : str
                    Filename.
    """
    with open(filename,'w',encoding='utf-8') as f:
        json.dump(settings,f,ensure_ascii=False, indent=4)
    return settings

def load_dictionary(filename):
    """
    Definition to load a dictionary (JSON) file.


    Parameters
    ----------
    filename      : str
                    Filename.

    Returns
    ----------
    settings      : dict
                    Dictionary read from the file.

    """
    settings = json.load(open(filename))
    return settings

def list_files(path,key='*.*',recursive=True):
    """
    Definition to list files in a given path with a given key.

    Parameters
    ----------
    path        : str
                  Path to a folder.
    key         : str
                  Key used for scanning a path.
    recursive   : bool
                  If set True, scan the path recursively.

    Returns
    ----------
    files_list  : ndarray
                  list of files found in a given path.
    """
    if recursive == True:
        search_result = pathlib.Path(path).rglob(key)
    elif recursive == False:
        search_result = pathlib.Path(path).glob(key)
    files_list = []
    for item in search_result:
        files_list.append(str(item))
    return files_list

def convert_bytes(num):
    """
    A definition to convert bytes to semantic scheme (MB,GB or alike). Inspired from https://stackoverflow.com/questions/2104080/how-can-i-check-file-size-in-python#2104083.

    Parameters
    ----------
    num        : float
                 Size in bytes

    Returns
    ----------
    num        : float
                 Size in new unit.
    x          : str
                 New unit bytes, KB, MB, GB or TB.
    """
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return num, x
        num /= 1024.0
    return None,None

def size_of_a_file(file_path):
    """
    A definition to get size of a file with a relevant unit.

    Parameters
    ----------
    file_path  : float
                 Path of the file.

    Returns
    ----------
    a          : float
                 Size of the file.
    b          : str
                 Unit of the size (bytes, KB, MB, GB or TB).
    """
    if os.path.isfile(file_path):
        file_info = os.stat(file_path)
        a,b       = convert_bytes(file_info.st_size)
        return a,b
    return None,None



from odak import np
from plyfile import PlyData, PlyElement

class PLY_object():
    """
    A class to create an PLY file. This is useful for visualizing rays in different CAD software such as FreeCAD, Blender or Meshlab.
    """
    def __init__(self):
        self.pnts = None
        self.tris = None
    def draw_a_ray(self,point0,point1,k=0.02,color=[255,255,255]):
        """
        Definition to draw a ray.

        Parameters
        ----------
        point0      : list
                      List that contains X, Y, and Z start locations of a ray.
        point1      : list
                      List that contains X, Y, and Z end locations of a ray.
        k           : float
                      Thickness of a ray.
        color       : list
                      List that contains red, green and blue color channel values (8 bit).
        """
        import numpy as np_cpu
        point0    = np_cpu.reshape(np_cpu.array(point0),(3))
        point1    = np_cpu.reshape(np_cpu.array(point1),(3))
        if type(self.pnts) == type(None):
            a = 0
        elif type(self.pnts) != type(None):
            a = np_cpu.asarray(self.pnts).shape[0]
        pnts      = np_cpu.array([
                                  (point0[0]   ,point0[1]   ,point0[2]),
                                  (point0[0]+k ,point0[1]   ,point0[2]),
                                  (point0[0]+k ,point0[1]+k ,point0[2]),
                                  (point0[0]   ,point0[1]+k ,point0[2]),
                                  (point1[0]   ,point1[1]   ,point1[2]),
                                  (point1[0]+k ,point1[1]   ,point1[2]),
                                  (point1[0]+k ,point1[1]+k ,point1[2]),
                                  (point1[0]   ,point1[1]+k ,point1[2])],
                                  dtype=[('x', 'f4'), ('y', 'f4'), ('z', 'f4')]
                                 )
        tris      = np_cpu.array([
                                  ([0+a,3+a,1+a], color[0], color[1], color[2]),
                                  ([1+a,3+a,2+a], color[0], color[1], color[2]),
                                  ([0+a,4+a,7+a], color[0], color[1], color[2]),
                                  ([0+a,7+a,3+a], color[0], color[1], color[2]),
                                  ([4+a,5+a,6+a], color[0], color[1], color[2]),
                                  ([4+a,6+a,7+a], color[0], color[1], color[2]),
                                  ([5+a,1+a,2+a], color[0], color[1], color[2]),
                                  ([5+a,2+a,6+a], color[0], color[1], color[2]),
                                  ([2+a,3+a,6+a], color[0], color[1], color[2]),
                                  ([3+a,7+a,6+a], color[0], color[1], color[2]),
                                  ([0+a,1+a,5+a], color[0], color[1], color[2]),
                                  ([0+a,5+a,4+a], color[0], color[1], color[2]) ],
                                  dtype=[('vertex_indices', 'i4', (3,)),
                                         ('red', 'u1'), ('green', 'u1'),
                                         ('blue', 'u1')]
                                )
        if type(self.pnts) != type(None):
            self.tris = np_cpu.concatenate([np.copy(self.tris),np.copy(tris)])
            self.pnts = np_cpu.concatenate([np.copy(self.pnts),np.copy(pnts)])
        else:
            self.tris = np_cpu.copy(tris)
            self.pnts = np_cpu.copy(pnts)

    def save_PLY(self,savefn='out.ply'):
        """
        Definition to save the PLY file.

        Parameters
        ----------
        savefn      : str
                      Filename with a complete path to save the PLY file.

        """
        el1       = PlyElement.describe(self.pnts, 'vertex', comments=['Vertex data'])
        el2       = PlyElement.describe(self.tris, 'face', comments=['Face data'])
        PlyData([el1,el2],text="True").write(savefn)
